//#define _CSRUTILS_INT_

#include <Windows.h>
#include <vector>

#include "csrmacro.h"
#include "csrmthread.h"

using namespace csrutils;

struct ThreadMonitor_ChildHolder {
	ThreadMonitor* monitor;
	double weight;
};

struct csrutils::ThreadMonitor_Job {
	double weight;
	volatile double percentage;
	std::vector<ThreadMonitor_ChildHolder> childs;
};

struct csrutils::ThreadMonitor_Ctx {
	std::vector<ThreadMonitor_Job> jobs;
	HANDLE hEventCompleted;
	DWORD mainThreadId;
	CRITICAL_SECTION criticalSection;
	volatile bool c_abort;
};

ThreadMonitor::ThreadMonitor() {
	ctx = new ThreadMonitor_Ctx;
	ThreadMonitor_Job job;
	job.weight     = 1.0;
	job.percentage = 0.0;
	ctx->jobs.push_back(job);
	ctx->hEventCompleted = CreateEvent(0,true,false,0);
	ctx->mainThreadId = GetCurrentThreadId();
	ctx->c_abort = false;
	InitializeCriticalSectionAndSpinCount(&ctx->criticalSection,100);
}

ThreadMonitor::~ThreadMonitor() {
	CloseHandle(ctx->hEventCompleted);
	DeleteCriticalSection(&ctx->criticalSection);
	delete ctx;
}

ThreadMonitor_Job& ThreadMonitor::getcurrentjob() {
	return *(ctx->jobs.end()-1);
}

bool ThreadMonitor::checkmainthread() {
	return ctx->mainThreadId == GetCurrentThreadId();
}

void ThreadMonitor::setmainthread() {
	ctx->mainThreadId = GetCurrentThreadId();
}

void ThreadMonitor::pushjob(double weight) {
	if( ! checkmainthread() )
		throw 0;
	ThreadMonitor_Job job;
	job.weight     = weight;
	job.percentage = 0.0;
	EnterCriticalSection(&ctx->criticalSection);
	ctx->jobs.push_back(job);
	LeaveCriticalSection(&ctx->criticalSection);
}

void ThreadMonitor::popjob() {
	if( ! checkmainthread() )
		throw 0;
	waitforchilds();
	EnterCriticalSection(&ctx->criticalSection);
	double w = getcurrentjob().weight;
	ctx->jobs.pop_back();
	getcurrentjob().percentage += w;
	LeaveCriticalSection(&ctx->criticalSection);
}

void ThreadMonitor::setjob(double per) {
	if( ! checkmainthread() )
		throw 0;
	getcurrentjob().percentage = per;
}

bool ThreadMonitor::checkabort() {
	if( ! checkmainthread() )
		throw 0;
	return ctx->c_abort;
}

void ThreadMonitor::setcompleted() {
	if( ! checkmainthread() )
		throw 0;
	waitforchilds();
	getcurrentjob().percentage = 1.0;
	if( ctx->jobs.size() == 1 ) {
		SetEvent(ctx->hEventCompleted);
	}
}

double ThreadMonitor::getpers() {
	double p = 0.0;
	EnterCriticalSection(&ctx->criticalSection);
	for( intmax_t i = ctx->jobs.size()-1; i>=0; i-- ) {
		ThreadMonitor_Job& job = ctx->jobs[i];
		for( auto iter = job.childs.begin(); iter != job.childs.end(); iter++ ) {
			ThreadMonitor_ChildHolder h = *iter;
			p += h.weight * h.monitor->getpers();
		}
		p = LIMIT(p+job.percentage,0.0,1.0);
		p *= job.weight;
	}
	LeaveCriticalSection(&ctx->criticalSection);
	return LIMIT(p,0.0,1.0);
}

void ThreadMonitor::cmd_abort() {
	ctx->c_abort = true;
}

void ThreadMonitor::cmd_pause() {
	// TODO:
}

void ThreadMonitor::cmd_unpause() {
	// TODO:
}

ThreadMonitorChild ThreadMonitor::createchild(ThreadMonitor_Proc proc, double weight, void* pUser) {
	if( ! checkmainthread() )
		throw 0;
	ThreadMonitor_ChildHolder h;
	h.weight  = weight;
	h.monitor = new ThreadMonitor();
	h.monitor->setuserdata(pUser);
	EnterCriticalSection(&ctx->criticalSection);
	getcurrentjob().childs.push_back(h);
	LeaveCriticalSection(&ctx->criticalSection);
	return ThreadMonitorChild( *h.monitor, proc );
}

void ThreadMonitor::waitforchilds() {
	if( ! checkmainthread() )
		throw 0;
	EnterCriticalSection(&ctx->criticalSection);
	ThreadMonitor_Job& job = getcurrentjob();
	size_t size = job.childs.size();
	HANDLE* events = new HANDLE[size];
	for( size_t i=0; i<size; i++ ) {
		ThreadMonitor_ChildHolder h = job.childs[i];
		events[i] = h.monitor->ctx->hEventCompleted;
	}
	LeaveCriticalSection(&ctx->criticalSection);
	WaitForMultipleObjects( (DWORD)size, events, true, (DWORD)(-1) );
	EnterCriticalSection(&ctx->criticalSection);
	for( auto iter = job.childs.begin(); iter != job.childs.end(); iter++ ) {
		ThreadMonitor_ChildHolder h = *iter;
		delete h.monitor;
	}
	job.childs.clear();
	LeaveCriticalSection(&ctx->criticalSection);
}

void ThreadMonitor::abortchilds() {
	if( ! checkmainthread() )
		throw 0;
	EnterCriticalSection(&ctx->criticalSection);
	ThreadMonitor_Job& job = getcurrentjob();
	for( auto iter = job.childs.begin(); iter != job.childs.end(); iter++ ) {
		ThreadMonitor_ChildHolder h = *iter;
		h.monitor->cmd_abort();
	}
	LeaveCriticalSection(&ctx->criticalSection);
	waitforchilds();
}

struct ThreadMonitorChild_RunCtx {
	ThreadMonitor* monitor;
	ThreadMonitor_Proc proc;
};

DWORD __stdcall ThreadMonitorChild_WinProc(LPVOID pArg) {
	ThreadMonitorChild_RunCtx* runctx = (ThreadMonitorChild_RunCtx*) pArg;
	runctx->monitor->setmainthread();
	runctx->proc( *runctx->monitor );
	runctx->monitor->setcompleted();
	delete runctx;
	return 0;
}

ThreadMonitorChild::ThreadMonitorChild(ThreadMonitor& _monitor, ThreadMonitor_Proc _proc)
		: monitor(_monitor), proc(_proc) {
}

void ThreadMonitorChild::run() {
	ThreadMonitorChild_RunCtx* runctx = new ThreadMonitorChild_RunCtx;
	runctx->monitor = &monitor;
	runctx->proc    = proc;
	CreateThread(0,0,ThreadMonitorChild_WinProc,runctx,0,0);
}

void ThreadMonitorChild::abort() {
	monitor.cmd_abort();
}

