/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **              DOM Store               **
 ******************************************/

#pragma once

#include "csrutils.h"
#include "csrstream.h"
#include "csrstr.h"

#include <map>
#include <new>
#include <vector>

namespace csrutils {

class DOMSObject;
class DOMSItem;
class DOMSIterator;
class DOMSConstIterator;

enum DOMSType {
	DOMS_NULL = 0,
	DOMS_INT = 1,
	DOMS_DOUBLE,
	DOMS_STRING,
	DOMS_OBJECT,
	DOMS_BINARY
};

class DOMSItemHolder {
public:
	DOMSType type;
	const StringB name;
	int number;
	bool use_name;
	size_t val_size;

	union {
		int val_int;
		double val_double;
		StringB* val_str;
		DOMSObject* val_obj;
		void* val_data;
	};

	//DOMSItemHolder() { };
	DOMSItemHolder(const DOMSItemHolder& h);
	DOMSItemHolder(const String& name);
	DOMSItemHolder(int number);
	~DOMSItemHolder();

	void set(int val);
	void set(double val);
	void set(const String& val);
	void set(const DOMSObject& val);
	void set(const void* data, size_t size);

	void clear();

	uint64_t size() const;
	uint64_t store(void* data, uint64_t avail) const;
	uint64_t load(void* data, uint64_t avail);
	static DOMSItemHolder* load2(void* data, uint64_t avail, uint64_t &size);
	void print(FILE* fd, int sublevel=0);
};

class DOMSObject {
	std::map<StringB,DOMSItemHolder*> params;
	std::vector<DOMSItemHolder*> items;

public:
	DOMSObject();
	DOMSObject(const DOMSObject& obj);
	~DOMSObject();

	//DOMSItem operator[] (const wchar_t* name);
	DOMSItem operator[] (const String& name);
	DOMSItem operator[] (int);
	DOMSIterator iter();
	DOMSConstIterator iter() const;

	uint64_t size() const;
	uint64_t store(void* data, uint64_t avail) const;
	uint64_t load(void* data, uint64_t avail);
	void print(FILE* fd, int sublevel=0);

protected:
	DOMSItemHolder& provideParam(const String& name);
	DOMSItemHolder& provideParam(int number);
	DOMSItemHolder* getParam(const String& name);
	DOMSItemHolder* getParam(int number);

	friend class DOMSItem;
	friend class DOMSIterator;
	friend class DOMSConstIterator;
};

class DOMSItem {
	DOMSObject& object;
	const StringB name;
	int number;
	bool use_name;

protected:
	DOMSItem(DOMSObject& object, const String& name);
	DOMSItem(DOMSObject& object, int number);
public:
	~DOMSItem();

	int operator= (int val);
	double operator= (double val);
	//const wchar_t* operator= (const wchar_t* val);
	const String& operator= (const String& val);
	const DOMSObject& operator= (const DOMSObject& val);
	void set_bin(const void* data, size_t size);

	operator int() const;
	operator double() const;
	//operator const char*() const;
	operator StringB() const;
	operator DOMSObject& () const;
	void* get_bin(size_t &size) const;

	void make_object();
	DOMSObject& obj();
	inline StringB str() const { return (StringB)*this; };

	bool exsists() const;
	bool isnumbered() const { return !use_name; }
	bool isnamed() const { return use_name; }
	const wchar_t* getname() const;
	const int getnumber() const;
	DOMSType type() const;
	/*bool is_int() const;
	bool is_double() const;
	bool is_string() const;
	bool is_object() const;
	bool is_binary() const;*/

	friend class DOMSObject;
	friend class DOMSIterator;
	friend class DOMSConstIterator;
};

class DOMSIterator {
	DOMSObject& object;

	typedef std::_Vector_iterator<std::_Vector_val<std::_Simple_types<csrutils::DOMSItemHolder *>>> _tveciter;
	_tveciter veciter;
	_tveciter vecbegin;
	_tveciter vecend;

	typedef std::_Tree_const_iterator<std::_Tree_val<std::_Tree_simple_types<std::pair<const StringB,DOMSItemHolder*>>>> _tmapiter;
	_tmapiter mapiter;
	_tmapiter mapbegin;
	_tmapiter mapend;

protected:
	DOMSIterator(DOMSObject& object);

public:
	void operator++ (int);
	DOMSItem operator* ();
	operator bool() const;
	void reset();
	bool has_next() const;

	friend class DOMSObject;
};


enum DOMSWriteFlags {
	DOMSWriteFlags_None = 0,
	DOMSWriteFlags_NullTerm
};

bool DOMSWrite(wchar_t*& data, size_t& size, DOMSObject& obj, DOMSWriteFlags flags);
bool DOMSRead(const wchar_t* data, size_t size, DOMSObject& obj, DOMSWriteFlags flags);
bool DOMSWriteFree(wchar_t* data, size_t size);

}