/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **       File functiond & utils         **
 ** Version 1.0 (2015 DECEMBER 04)       **
 ******************************************/

#pragma once

#define _CSRUTILS_FILE_

#include <csrutils.h>
#include <csrstr.h>
#include <csrstream.h>
#include <csrtime.h>
#include <csrurl.h>

namespace csrutils {

class Url;
class File;
class FileBase;
class FileIterator;
class FileIteratorBase;
class FileHandle;
//class FileInfo;

struct FileCtx;
struct FileIteratorCtx;
struct FileHandleCtx;

struct FILE_INFO;

enum FileType {
	FileType_Null = 0,
	FileType_File,
	FileType_Directory,
	FileType_Virtual
};

struct FileMD5 {
	uint8_t digest[16];

	inline FileMD5() {
		*((uint64_t*)digest+0) = 0;
		*((uint64_t*)digest+1) = 0;
	}

	inline FileMD5(uint8_t _digest[16]) {
		*((uint64_t*)digest+0) = *((uint64_t*)_digest+0);
		*((uint64_t*)digest+1) = *((uint64_t*)_digest+1);
	}

	inline FileMD5(const FileMD5& m) {
		*((uint64_t*)digest+0) = *((uint64_t*)m.digest+0);
		*((uint64_t*)digest+1) = *((uint64_t*)m.digest+1);
	}

	inline FileMD5& operator= (const FileMD5& m) {
		*((uint64_t*)digest+0) = *((uint64_t*)m.digest+0);
		*((uint64_t*)digest+1) = *((uint64_t*)m.digest+1);
		return *this;
	}

	inline bool operator== (const FileMD5& m) const {
		return *((uint64_t*)digest+0) == *((uint64_t*)m.digest+0)
			&& *((uint64_t*)digest+1) == *((uint64_t*)m.digest+1);
	}

	operator StringB() const;
};

class FileBase {
	int guid = 0;

public:
	virtual bool getexists() = 0;
	virtual FileType gettype() = 0;
	virtual uint64_t getsize() = 0;
	virtual FileMD5 md5() = 0;

	virtual FileHandle* open(uint32_t flags) = 0;
	virtual FileIteratorBase* getfiles() = 0;
	virtual FileBase* getchild(String path) = 0;
	virtual StringB getname() = 0;
	virtual void info(FILE_INFO&) = 0;
	virtual Url getpath() = 0;
	virtual FileBase* getfile(const Url& path) = 0;

	virtual bool remove() = 0;
	virtual bool copy(const File& dst);
	virtual bool createdir() = 0;
	virtual bool createpath();

	virtual FileBase* clone() = 0;
};

class File {
	//FileCtx* ctx;
	FileBase* base;

public:
	//File();
	File(FileBase* base);
	File(const File& file);
	File(String path);
	File(const File& base, String path);
	~File();

	inline bool exists() const { return base->getexists(); }
	inline FileType type() const { return base->gettype(); }
	inline uint64_t size() const { return base->getsize(); }
	inline FileMD5 md5() const { return base->md5(); }
	FileIterator files() const;
	StringB name() const;
	inline Url path() const { return base->getpath(); }
	inline File getfile(const Url& path) const { return File(base->getfile(path)); }
	File child(String path) const;
	inline FileBase* getbase() const { return base; }
	inline void info(FILE_INFO& inf) { base->info(inf); }
	inline FileHandle* open(uint32_t flags) const { return base->open(flags); }

	inline bool remove() { return base->remove(); }
	inline bool copy(const File& dst) { return base->copy(dst); }
	inline bool createdir() { return base->createdir(); }
	inline bool createpath() { return base->createpath(); }

	static File getcurdir();
};

class FileIteratorBase {
public:
	virtual FileBase* getnext() = 0;
	virtual void begin() = 0;
};

class FileIterator {
	FileIteratorCtx* ctx;
	FileIteratorBase* base;
	FileBase* curfile;

public:
	FileIterator(const File& dir);
	~FileIterator();

	operator File() const;
	operator bool() const;
	inline File operator *() const { return (File)*this; }
	void operator ++(int);
};

class FilePointer {
	FileHandle &handle;
public:
	FilePointer(FileHandle &_handle) : handle(_handle) { }
	void operator+=(size_t off);
	void operator-=(size_t off);
	operator size_t();
	FilePointer& operator=(size_t pos);
};

const uint32_t File_Read = 0x01;
const uint32_t File_Write = 0x02;

class FileHandle : public FileStream {
public:
/*
Read 'size' bytes from input stream into 'data' buffer.
Number of read bytes stored int 'read' variable.
In case of an error returns non-zero error and 'read'
is number of successfuly read bytes before error eccured.
Error code doesn't mean that no data has been read, 'read'=0 does.
Condition 'read' < 'size' doesn't indicate any error.
End of input stream indicated by returning IO_EOF.
*/
	virtual int read(void *data, size_t size, size_t& read) = 0;

/*
Write 'size' bytes into output stream from 'data' buffer.
Number of written bytes stored int 'written' variable.
In case of an error returns non-zero error and 'written'
is number of successfuly written bytes before error eccured.
Error code doesn't mean that no data has been written, 'written'=0 does.
Condition 'written' < 'size' doesn't indicate any error.
*/
	virtual int write(const void *data, size_t size, size_t& written) = 0;

/*
Get size of file. In case of an error -1 is returned.
*/
	virtual uint64_t getsize() = 0;

/*
Get offset of file pointer relative to file begining.
To be used with setposition() function.
In case of an error -1 is returned.
*/
	virtual uint64_t getposition() = 0;

/*
Set offset of file pointer relative to file begining.
To be used with getposition() function.
In case of an error -1 is returned.
*/
	virtual bool setposition(uint64_t pos) = 0;

/*
Truncate file to current pointer position.
*/
	virtual bool setend() = 0;

	FilePointer pos() { return FilePointer(*this); }

	inline void close() {
		heap_free(this);
	}
};

class FileHandleDisk : public FileHandle {
	FileHandleCtx* ctx;

public:
	FileHandleDisk(File file, uint32_t flags);
	~FileHandleDisk();

	int read(void *data, size_t size, size_t& read);
	int write(const void *data, size_t size, size_t& written);

	uint64_t getsize();
	uint64_t getposition();
	bool setposition(uint64_t pos);
	bool setend();
};

struct FILE_INFO {
	uint64_t fiSize;
	Time flCreated;
	Time flAccessed;
	Time flModified;
};

inline void FilePointer::operator+=(size_t off) {
	handle.setposition( handle.getposition() + off );
}
inline void FilePointer::operator-=(size_t off) {
	handle.setposition( handle.getposition() - off );
}
inline FilePointer::operator size_t() {
	return handle.getposition();
}
inline FilePointer& FilePointer::operator=(size_t pos) {
	handle.setposition(pos);
	return *this;
}

}
