/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **                 File                 **
 ** Version 0.0 (2015 AUGUST 06)         **
 ******************************************/

#include <exception>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <dirent.h>

#include <csrfile.h>
#include <csrfile_int.h>
#include <md5.h>

using namespace csrutils;


class Error_FileNotExists : public std::exception {
};


struct csrutils::FileHandleCtx {
	int fd;
	size_t size;
	uint32_t flags;
};


FileBaseDisk::FileBaseDisk(String path) {
	//ctx = new FileBaseDisk_Ctx;
	b_path = path;
	u_path = StringAB(path,STRING_UTF8);
}

FileBaseDisk::FileBaseDisk(const FileBaseDisk& base) {
	//ctx = new FileBaseDisk_Ctx;
	b_path = base.b_path;
	u_path = base.u_path;
}

FileBaseDisk::~FileBaseDisk() {
	//delete ctx;
}

bool FileBaseDisk::getexists() {
	struct stat64 st;
	return stat64(u_path.bytes(),&st) == 0;
}

FileType FileBaseDisk::gettype() {
	struct stat64 st;
	if( stat64(u_path.bytes(),&st) != 0 )
		return FileType_Null;
	if( S_ISDIR(st.st_mode) )
		return FileType_Directory;
	if( S_ISREG(st.st_mode) )
		return FileType_File;
	return FileType_Null;
}

uint64_t FileBaseDisk::getsize() {
	struct stat64 st;
	if( stat64(u_path.bytes(),&st) != 0 )
		return -1;
	return st.st_size;
}

FileMD5 FileBaseDisk::md5() {
	md5_state_t md5s;
	md5_init(&md5s);
	int fd = ::open( u_path.bytes(), O_RDONLY|O_LARGEFILE );
	if( fd==-1 )
		return FileMD5();
	size_t sz = 0;
	char* buf = new char[4096];
	do {
		sz = read(fd,buf,4096);
		md5_append(&md5s,(md5_byte_t*)buf,sz);
	} while( sz>0 );
	delete[] buf;
	FileMD5 md5d;
	md5_finish(&md5s,md5d.digest);
	return md5d;
}

FileHandle* FileBaseDisk::open(uint32_t flags) {
	FileHandle* fh;
	try {
		fh = new FileHandleDisk(File(this),flags);
	} catch( Error_FileNotExists e ) {
		fh = NULL;
	}
	return fh;
}

Url FileBaseDisk::getpath() {
	return b_path;
}

FileBase* FileBaseDisk::getfile(const Url& path) {
	return new FileBaseDisk(path);
}

bool FileBaseDisk::remove() {
	if( !getexists() )
		return false;
	return ::remove( u_path.bytes() ) == 0;
	// TODO: remove directory feature
}

bool FileBaseDisk::createdir() {
	if( getexists() )
		return false;
	return mkdir( u_path.bytes(), S_IRWXU|S_IRWXG|S_IROTH|S_IXOTH) == 0;
}

FileIteratorBase* FileBaseDisk::getfiles() {
	return new FileIteratorBaseDisk(this);
}

FileBase* FileBaseDisk::getchild(String path) {
	return new FileBaseDisk( Url(b_path+L"\\"+path) );
}

StringB FileBaseDisk::getname() {
	return Url(b_path).name();
}

void FileBaseDisk::info(FILE_INFO& info) {
	struct stat64 st;
	if( stat64(u_path.bytes(),&st) != 0 )
		throw 0;
	info.fiSize = st.st_size;
	// TODO:
	//info.flCreated	= Time::fromUnix(st.st_ctime);
	//info.flModified = Time::fromUnix(st.st_mtime);
	//info.flAccessed = Time::fromUnix(st.st_atime);
}

FileBase* FileBaseDisk::clone() {
	return new FileBaseDisk(*this);
}

FileIteratorBaseDisk::FileIteratorBaseDisk(FileBase* directory) {
	dir = (FileBaseDisk*) directory->clone();
	dp = NULL;
	begin();
}

FileIteratorBaseDisk::~FileIteratorBaseDisk() {
	closedir(dp);
}

void FileIteratorBaseDisk::begin() {
	if( dp != NULL )
		closedir(dp);
	dp = opendir( dir->u_path.bytes() );
}

FileBase* FileIteratorBaseDisk::getnext() {
	struct dirent *ep = readdir(dp);
	if( ep==NULL )
		return NULL;
	if( strcmp(".",ep->d_name) == 0 )
		return getnext();
	if( strcmp("..",ep->d_name) == 0 )
		return getnext();
	FileBase* file = dir->getchild( StringB(StringA(ep->d_name,STRING_UTF8)) );
	return file;
}


FileHandleDisk::FileHandleDisk(File file, uint32_t flags) {
	ctx = new FileHandleCtx;
	ctx->flags = flags;
	StringAB path0( StringB(file.path()), STRING_UTF8);
	char *path = path0.bytes();
	
	if( ( flags & (File_Write|File_Read) ) == (File_Write|File_Read) )
		ctx->fd = open(path,O_RDWR|O_CREAT|O_LARGEFILE,S_IRWXU|S_IRWXG|S_IROTH);
	else if( ( flags & (File_Write|File_Read) ) == File_Write )
		ctx->fd = open(path,O_WRONLY|O_CREAT|O_TRUNC|O_LARGEFILE,S_IRWXU|S_IRWXG|S_IROTH);
	else if( ( flags & (File_Write|File_Read) ) == File_Read )
		ctx->fd = open(path,O_RDONLY|O_LARGEFILE);
	else
		ctx->fd = -1;
	if( ctx->fd==-1 )
		throw Error_FileNotExists();
	struct stat64 st;
	fstat64(ctx->fd,&st);
	ctx->size = st.st_size;
}

FileHandleDisk::~FileHandleDisk() {
	if( ctx->fd!=-1 )
		::close(ctx->fd);
	delete ctx;
}

uint64_t FileHandleDisk::getsize() {
	return ctx->size;
}

uint64_t FileHandleDisk::getposition() {
	return lseek64(ctx->fd,0,SEEK_CUR);
}

bool FileHandleDisk::setposition(uint64_t pos) {
	return lseek64(ctx->fd,pos,SEEK_SET) != -1;
}

bool FileHandleDisk::setend() {
	return ftruncate64(ctx->fd,getposition()) == 0;
}

int FileHandleDisk::read(void *data, size_t size, size_t& readed) {
	if( ! (ctx->flags & File_Read) ) {
		readed = 0;
		return IO_ERROR;
	}
	size_t szrd = ::read(ctx->fd,data,size);
	readed = szrd;
	if( szrd==0 && size!=0 ) {
		readed = 0;
		return IO_EOF;
	}
	return 0;
}

int FileHandleDisk::write(const void *data, size_t size, size_t& written) {
	if( ! (ctx->flags & File_Write) ) {
		written = 0;
		return IO_ERROR;
	}
	size_t szwr = ::write(ctx->fd,data,size);
	written = szwr;
	if( szwr==0 && size!=0 ) {
		written = 0;
		return IO_ERROR;
	}
	return 0;
}
