/******************************************
 **              CSR UTILS               **
 ** http://info.dcsr.ru/comsp/csrutils/  **
 ******************************************
 **                 File                 **
 ** Version 1.0 (2015 DECEMBER 04)       **
 ******************************************/

#define _CRT_NON_CONFORMING_SWPRINTFS
#define _CRT_SECURE_NO_WARNINGS

#include <exception>
#include <Windows.h>
#include <Shlwapi.h>
#include <stdio.h>

#include "csrfile.h"
#include "csrfile_int.h"
#include "md5.h"

using namespace csrutils;

/*struct csrutils::FileBaseDisk_Ctx {
	HANDLE hFile;
	FileType type;
};
*/

inline FileType Win32_GetFileType(const wchar_t* path) {
	DWORD attr = GetFileAttributesW(path);
	if( attr == INVALID_FILE_ATTRIBUTES )
		return FileType_Null;
	if( attr & FILE_ATTRIBUTE_DIRECTORY )
		return FileType_Directory;
	return FileType_File;
}

class Error_FileNotExists : public std::exception {
};


struct csrutils::FileHandleCtx {
	HANDLE hFile;
	size_t size;
	uint32_t flags;
};

FileBaseDisk::FileBaseDisk(String path) {
	//ctx = new FileBaseDisk_Ctx;
	b_path = path;
	b_type = Win32_GetFileType(b_path.bytes());
}

FileBaseDisk::FileBaseDisk(const FileBaseDisk& base) {
	//ctx = new FileBaseDisk_Ctx;
	b_path = base.b_path;
	b_type = base.b_type;
}

FileBaseDisk::~FileBaseDisk() {
}

bool FileBaseDisk::getexists() {
	return PathFileExistsW( b_path.bytes() ) == TRUE;
}

FileType FileBaseDisk::gettype() {
	//b_type = Win32_GetFileType(b_path.bytes());
	return b_type;
}

uint64_t FileBaseDisk::getsize() {
	WIN32_FILE_ATTRIBUTE_DATA fad;
    if( ! GetFileAttributesExW( b_path.bytes(), GetFileExInfoStandard, &fad) )
        return -1;
	return ( (uint64_t)fad.nFileSizeHigh << 32 ) | fad.nFileSizeLow;
}

FileMD5 FileBaseDisk::md5() {
	md5_state_t md5s;
	md5_init(&md5s);
	HANDLE hFile = CreateFileW( b_path.bytes(), GENERIC_READ,
			FILE_SHARE_READ, 0, OPEN_ALWAYS, 0, 0 );
	if( hFile==INVALID_HANDLE_VALUE )
		return FileMD5();
	DWORD sz = 0;
	char* buf = heap_alloc<char>(4096);
	do {
		ReadFile(hFile,buf,4096,&sz,0);
		md5_append(&md5s,(md5_byte_t*)buf,4096);
	} while( sz>0 );
	heap_free(buf);
	FileMD5 md5d;
	md5_finish(&md5s,md5d.digest);
	return md5d;
}

FileHandle* FileBaseDisk::open(uint32_t flags) {
	FileHandle* fh;
	try {
		fh = new(heap_alloc<FileHandleDisk>()) FileHandleDisk(File(this),flags);
	} catch( Error_FileNotExists e ) {
		fh = NULL;
	}
	return fh;
}

Url FileBaseDisk::getpath() {
	return Url( URL_WINDOWS | URL_ABSOLUTE
			| (b_type==FileType_Directory?URL_DIRECTORY:0),
		b_path);
}

FileBase* FileBaseDisk::getfile(const Url& path) {
	return new(heap_alloc<FileBaseDisk>()) FileBaseDisk(path);
}

bool FileBaseDisk::remove() {
	if( !getexists() )
		return false;
	return TRUE == DeleteFileW( b_path.bytes() );
}

bool FileBaseDisk::createdir() {
	if( getexists() )
		return false;
	if( TRUE == CreateDirectoryW( b_path.bytes(), NULL ) ) {
		b_type = FileType_Directory;
		return true;
	} else
		return false;
}

FileIteratorBase* FileBaseDisk::getfiles() {
	return new(heap_alloc<FileIteratorBaseDisk>()) FileIteratorBaseDisk(this);
}

FileBase* FileBaseDisk::getchild(String path) {
	return new(heap_alloc<FileBaseDisk>()) FileBaseDisk( Url(b_path+L"\\"+path) );
}

StringB FileBaseDisk::getname() {
	return Url(b_path).name();
}

void FileBaseDisk::info(FILE_INFO& info) {
	WIN32_FILE_ATTRIBUTE_DATA fad;
    if( ! GetFileAttributesExW( b_path.bytes(), GetFileExInfoStandard, &fad) )
        throw 0;
	info.fiSize = ( (uint64_t)fad.nFileSizeHigh << 32 ) | fad.nFileSizeLow;
	info.flCreated	= Time::fromWindows(fad.ftCreationTime);
	info.flModified = Time::fromWindows(fad.ftLastWriteTime);
	info.flAccessed = Time::fromWindows(fad.ftLastAccessTime);
}

FileBase* FileBaseDisk::clone() {
	return new(heap_alloc<FileBaseDisk>()) FileBaseDisk(*this);
}

FileIteratorBaseDisk::FileIteratorBaseDisk(FileBase* directory) {
	dir = (FileBaseDisk*) directory->clone();
	hFind = NULL;
	begin();
}

FileIteratorBaseDisk::~FileIteratorBaseDisk() {
	FindClose(hFind);
}

void FileIteratorBaseDisk::begin() {
	if( hFind != NULL )
		FindClose(hFind);
	String dirpath = dir->b_path;
	wchar_t* buf = heap_alloc<wchar_t>( dirpath.len()+3 );
	swprintf( buf, L"%s\\*", dirpath.bytes() );
	hFind = FindFirstFileW(buf,&findData);
	heap_free(buf);
	if( hFind == INVALID_HANDLE_VALUE ) {
		found = false;
		return;
	}
	while( found && ( wcscmp(findData.cFileName,L".") == 0
			|| wcscmp(findData.cFileName,L"..") == 0 ) )
		found = FindNextFileW(hFind,&findData) != 0;
}

FileBase* FileIteratorBaseDisk::getnext() {
	if( !found )
		return NULL;
	FileBase* file = dir->getchild( findData.cFileName );
	found = FindNextFileW(hFind,&findData) != 0;
	return file;
}


FileHandleDisk::FileHandleDisk(File file, uint32_t flags) {
	ctx = heap_alloc<FileHandleCtx>();
	ctx->flags = flags;
	StringB path = file.path();
	if( ( flags & (File_Write|File_Read) ) == (File_Write|File_Read) )
		ctx->hFile = CreateFileW( path.bytes(), GENERIC_WRITE|GENERIC_READ,
			0, 0, OPEN_ALWAYS, 0, 0 );
	else if( ( flags & (File_Write|File_Read) ) == File_Write )
		ctx->hFile = CreateFileW( path.bytes(), GENERIC_WRITE,
			0, 0, CREATE_ALWAYS, 0, 0 );
	else if( ( flags & (File_Write|File_Read) ) == File_Read )
		ctx->hFile = CreateFileW( path.bytes(), GENERIC_READ,
			FILE_SHARE_READ, 0, OPEN_EXISTING, 0, 0 );
	else
		ctx->hFile = (HANDLE)-1;
	if( ctx->hFile==(HANDLE)-1 )
		throw Error_FileNotExists();
	GetFileSizeEx( ctx->hFile, (PLARGE_INTEGER) &ctx->size );
}

FileHandleDisk::~FileHandleDisk() {
	if( ctx->hFile != NULL )
		CloseHandle(ctx->hFile);
	heap_free(ctx);
}

uint64_t FileHandleDisk::getsize() {
	return ctx->size;
}

uint64_t FileHandleDisk::getposition() {
	LARGE_INTEGER pos0, pos1;
	pos0.QuadPart = 0;
	SetFilePointerEx(ctx->hFile,pos0,&pos1,FILE_CURRENT);
	return pos1.QuadPart;
}

bool FileHandleDisk::setposition(uint64_t pos) {
	LARGE_INTEGER pos0, pos1;
	pos0.QuadPart = pos;
	return TRUE == SetFilePointerEx(ctx->hFile,pos0,&pos1,FILE_BEGIN);
}

bool FileHandleDisk::setend() {
	return TRUE == SetEndOfFile(ctx->hFile);
}

int FileHandleDisk::read(void *data, size_t size, size_t& readed) {
	if( ! (ctx->flags & File_Read) )
		return IO_ERROR;
	DWORD szrd;
	ReadFile(ctx->hFile,data,(DWORD)size,&szrd,0);
	readed = szrd;
	if( szrd==0 && size!=0 )
		return IO_EOF;
	return 0;
}

int FileHandleDisk::write(const void *data, size_t size, size_t& written) {
	if( ! (ctx->flags & File_Write) )
		return IO_ERROR;
	DWORD szwr;
	WriteFile(ctx->hFile,data,(DWORD)size,&szwr,0);
	written = szwr;
	if( szwr==0 && size!=0 )
		return IO_ERROR;
	return 0;
}

/*
uint64_t FileHandle::getsize();
uint64_t FileHandle::getposition();
bool FileHandle::setposition(uint64_t pos);*/
