/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **           Heap functions             **
 ** Version 1.1 (2016 APRIL 09)          **
 ******************************************/

#pragma once

#include <csrutils.h>
#include <csrmem.h>

namespace csrutils {

const size_t HEAP_STRICT	= 0x0001;		// do not use 'new'/'malloc'
const size_t DHEAP_CACHE	= 0x0002;		// do not store volatile data in main memory (e.g. for file caches)
const size_t DHEAP_LOCKS	= 0x0004;		// use lock/unlock and usage counters
const size_t DHEAP_STATIC	= 0x0008;		// do not move blocks in memory
const size_t DHEAP_KEEP		= 0x0010;		// keep pages in memory (don't release them)
const size_t DHEAP_ALIGN16	= 0x0100;		// align by 16 bytes;
const size_t DHEAP_ALIGN32	= 0x0200;		// align by 32 bytes;

/* Notes:
	When using multitheading either DHEAP_STATIC or DHEAP_LOCKS should be active,
	otherwise it might cause pointers/data corruption.
*/

struct Heap_Ctx;

CSRLIB Heap_Ctx* heap_create(uint64_t flags);
CSRLIB void heap_destroy(Heap_Ctx* ctx);
CSRLIB void* heap_alloc(Heap_Ctx* ctx, size_t size);
CSRLIB void heap_free(Heap_Ctx* ctx, void* mem);
CSRLIB void* heap_resize(Heap_Ctx* ctx, void* mem, size_t size);

class Heap {
	Heap_Ctx *ctx;

public:
	inline Heap(Heap_Ctx *ctx) : ctx(ctx) { }

	inline Heap(uint64_t flags) { ctx = heap_create(flags); }

	//inline ~Heap() {
	//	heap_destroy(ctx);
	//}

	inline void destroy() { heap_destroy(ctx); }

	inline Heap_Ctx* getctx() const { return ctx; }

	inline void *alloc(size_t size) {
		return heap_alloc(ctx,size);
	}
	
	inline void free(void* mem) {
		return heap_free(ctx,mem);
	}
	
	void* resize(void* mem, size_t size) {
		return heap_resize(ctx,mem,size);
	}

	template<typename T> T* alloc() {
		return (T*) alloc( sizeof(T) );
	}

	template<typename T> T* alloc(size_t count) {
		return (T*) alloc( sizeof(T)*count );
	}

	template<typename T> void free(T* mem) {
		mem->~T();
		free( (void*)mem );
	}

	template<typename T> void free(T* mem, size_t count) {
		for( size_t i=0; i<count; i++ )
			(mem+i)->~T();
		free( (void*)mem );
	}
};

}
