/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **             Image utils              **
 ** Version: 0.2 (2015-05-25)            **
 ******************************************/

#pragma once

#define _CSRUTILS_IMAGE_

#include <csrutils.h>
#include <csrfile.h>

#ifdef _CSRMACRO_INC_
#define STRIDELENGTH24(width)	ALIGN32((width)*3)
#define STRIDELENGTH8(width)	ALIGN32((width)*1)
#endif

namespace csrutils {

class Image;
class ImageBase;

// Image bit pixel format
enum ImageBitFormat {
	ImageBitFormat_RGB24 = 1,
	ImageBitFormat_GRAY8
};

// Image file format
enum ImageFileFormat {
	ImageFileFormat_BMP = 1,
	ImageFileFormat_JPEG,
	ImageFileFormat_PNG
};

const uint32_t ImageEncoderParams_Quality = 0x0001;

// Image ecnoder parameters
struct ImageEncoderParams {
	uint32_t mask;		// mask of set parameters
	int Quality;		// compression quality, mask bit: ImageEncoderParams_Quality
};

struct Image_Ctx;

class ImageBase : public HandledObjectSync {
	Image_Ctx* ctx;
	ImageBitFormat format;
	uint32_t width;
	uint32_t height;
	uint8_t* data;

public:
	ImageBase(const ImageBase& img);
	ImageBase(ImageBase&& img);
	ImageBase(uint32_t width, uint32_t height, ImageBitFormat fmt);
#ifdef _WINDOWS_
	ImageBase(uint32_t width, uint32_t height, ImageBitFormat fmt, HDC hDC, HBITMAP& hBitmap);
#endif
	~ImageBase();

	inline uint32_t getWidth() const { return width; }
	inline uint32_t getHeight() const { return height; }
	inline const uint8_t* getData() const { return data; }
	inline uint8_t* getData() { return data; }

	ImageBase& operator=(const ImageBase& img);
	ImageBase& operator=(ImageBase&& img);

	void setPixel(uint32_t x, uint32_t y, uint32_t col);
	uint32_t getPixel(uint32_t x, uint32_t y) const;
	size_t getStrideSize() const;

	bool save(const wchar_t* path, ImageFileFormat fmt, ImageEncoderParams* params=0) const;
	static ImageBase* load(const wchar_t* path, ImageFileFormat fmt);
#ifdef _CSRUTILS_FILE_
	bool save(const File& file, ImageFileFormat fmt) const;
	static ImageBase* load(const File& file, ImageFileFormat fmt);
#endif

	friend class Image;
};

/*
Image class
Class provides interface for manipulating images/bitmaps.
It holds a pointer to ImageBase with uses counter,
there is no need to explicitly delete image,
deconstructed will handle memory release when it's needed.
Expression `Image new_img = img` creates a reference for `img` not a copy,
to create a full copy use `Image copy_img = img.clone()`.
*/
class Image {
	ImageBase* base;

public:
	inline Image() : base(NULL) { }
	inline Image(ImageBase *_base) : base(_base) {
		base->addHandle();
	}
	inline Image(const Image &img) : base(img.base) {
		base->addHandle();
	}
	inline Image(Image &&img) : base(img.base) {
		img.base = NULL;
	}
	inline Image(uint32_t width, uint32_t height, ImageBitFormat fmt) {
		base = new ImageBase(width,height,fmt);
		base->addHandle();
	}
#ifdef _WINDOWS_
	inline Image(uint32_t width, uint32_t height, ImageBitFormat fmt, HDC hDC, HBITMAP& hBitmap) {
		base = new ImageBase(width,height,fmt,hDC,hBitmap);
		base->addHandle();
	}
#endif
	inline ~Image() {
		if( base != NULL )
			base->delHandle();
	}

	inline Image& operator=(const Image &img) {
		if( base!=NULL )
			base->delHandle();
		base = img.base;
		base->addHandle();
		return *this;
	}

	inline Image& operator=(Image &&img) {
		if( base!=NULL )
			base->delHandle();
		base = img.base;
		img.base = NULL;
		return *this;
	}

	inline Image clone() const {
		ImageBase* newbase = new ImageBase(*base);
		Image img;
		newbase->addHandle();
		img.base = newbase;
		return img;
	}

	inline uint32_t getWidth() const { return base!=NULL ? base->getWidth() : 0; }
	inline uint32_t getHeight() const { return base!=NULL ? base->getHeight() : 0; }
	inline ImageBitFormat getFormat() const { return base!=NULL ? base->format : (ImageBitFormat)0; }
	inline const uint8_t* getData() const { return base!=NULL ? base->getData() : NULL; }
	inline uint8_t* getData() { return base!=NULL ? base->getData() : NULL; }
	inline void setPixel(uint32_t x, uint32_t y, uint32_t col) { base->setPixel(x,y,col); }
	inline uint32_t getPixel(uint32_t x, uint32_t y) const { return base->getPixel(x,y); }
	inline size_t getStrideSize() const { return base->getStrideSize(); }
	inline bool save(const wchar_t* path, ImageFileFormat fmt, ImageEncoderParams* params=0) const {
		return base->save(path,fmt,params);
	}
	inline static Image _load(const wchar_t* path, ImageFileFormat fmt) {
		return Image( ImageBase::load(path,fmt) );
	}
#ifdef _CSRUTILS_FILE_
	inline bool save(const File& file, ImageFileFormat fmt) const {
		return base->save(file,fmt);
	}
	inline bool load(const File& file, ImageFileFormat fmt) {
		ImageBase* newbase = ImageBase::load(file,fmt);
		if( newbase==NULL )
			return false;
		if( base!=NULL )
			base->delHandle();
		base = newbase;
		base->addHandle();
		return true;
	}
	inline static Image _load(const File& file, ImageFileFormat fmt) {
		return Image( ImageBase::load(file,fmt) );
	}
#endif
};

};