/******************************************
 **              CSR UTILS               **
 ** http://info.dcsr.ru/comp/csrutils/   **
 ******************************************
 **             Image utils              **
 **   Version: 0.2 (2015-05-25)          **
 ******************************************/

#include <Windows.h>
#include <gdiplus.h>

#include "csrmacro.h"
#include "csrfile.h"
#include "csrimage.h"

#pragma comment(lib,"Gdiplus.lib")

using namespace csrutils;

struct csrutils::Image_Ctx {
	HBITMAP hBitmap;
};

bool Gdiplus_Inited = false;

void Gdiplus_Init() {
	if(Gdiplus_Inited)
		return;
	Gdiplus::GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;
	Gdiplus::GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);
	Gdiplus_Inited = true;
}

int GetEncoderClsid(const WCHAR* format, CLSID* pClsid) {
	UINT num = 0;
	UINT size = 0;
	Gdiplus::ImageCodecInfo* pImageCodecInfo = NULL;
	Gdiplus::GetImageEncodersSize(&num, &size);
	if(size == 0)
		return -1;
	pImageCodecInfo = (Gdiplus::ImageCodecInfo*)(malloc(size));
	if(pImageCodecInfo == NULL)
		return -1;
	GetImageEncoders(num, size, pImageCodecInfo);
	for( UINT j = 0; j < num; ++j ) {
		if( wcscmp(pImageCodecInfo[j].MimeType, format) == 0 ) {
			*pClsid = pImageCodecInfo[j].Clsid;
			free(pImageCodecInfo);
			return j;
		}    
	}
	free(pImageCodecInfo);
	return -1;
}

ImageBase::ImageBase(const ImageBase &img) {
	ctx = new Image_Ctx;
	ctx->hBitmap = 0;
	format = img.format;
	width  = img.width;
	height = img.height;
	size_t size = STRIDELENGTH24(width)*height;
	data = (uint8_t*) VirtualAlloc(0,size,MEM_COMMIT,PAGE_READWRITE);
	memcpy(data,img.data,size);
}

ImageBase::ImageBase(ImageBase &&img) {
	ctx = img.ctx;
	img.ctx = 0;
	format = img.format;
	width  = img.width;
	height = img.height;
	data   = img.data;
}

ImageBase::ImageBase(uint32_t _width, uint32_t _height, ImageBitFormat _fmt)
		: width(_width), height(_height), format(_fmt) {
	ctx = new Image_Ctx;
	ctx->hBitmap = 0;
	size_t size = getStrideSize()*height;
	data = (uint8_t*) VirtualAlloc(0,size,MEM_COMMIT,PAGE_READWRITE);
}

ImageBase::ImageBase(uint32_t _width, uint32_t _height, ImageBitFormat _fmt, HDC hDC, HBITMAP& hBitmap)
		: width(_width), height(_height), format(_fmt) {
	if(format!=ImageBitFormat_RGB24)
		throw 0;
	ctx = new Image_Ctx;
	size_t size = STRIDELENGTH24(width)*height;
	BITMAPINFOHEADER bmih = {0};
	bmih.biSize				= sizeof(BITMAPINFOHEADER);
    bmih.biWidth			= width;
    bmih.biHeight			= height;
    bmih.biPlanes			= 1;
    bmih.biBitCount			= 24;
    bmih.biCompression		= 0;
    bmih.biSizeImage		= (DWORD) size;
    bmih.biXPelsPerMeter	= 0;
    bmih.biYPelsPerMeter	= 0;
    bmih.biClrUsed			= 0;
    bmih.biClrImportant		= 0;
	BITMAPINFO bmi = { bmih, 0 };
	ctx->hBitmap = CreateDIBSection(hDC,&bmi,DIB_RGB_COLORS,(void**)&data,0,0);
	hBitmap = ctx->hBitmap;
}

ImageBase::~ImageBase() {
	//delete[] data;
	if( ctx==NULL )
		return;
	if( ctx->hBitmap != NULL ) {
		DeleteObject(ctx->hBitmap);
	} else {
		size_t size = STRIDELENGTH24(width)*height;
		VirtualFree(data,size,MEM_RELEASE);
	}
	delete ctx;
}

uint32_t ImageBase::getPixel(uint32_t x, uint32_t y) const {
	if(format!=ImageBitFormat_RGB24)
		throw 0;
	if( x<0 || y<0 || x>=width || y>=height )
		return 0;
	size_t offset = 3*x+(height-y-1)*STRIDELENGTH24(width);
	if( offset+4 >= height*STRIDELENGTH24(width) )
		return ( 0xFFFFFF00 & ( *(uint32_t*) (data+offset-1) ) )>>8;
	else
		return 0xFFFFFF & ( *(uint32_t*) (data+offset) );
}

size_t ImageBase::getStrideSize() const {
	switch(format) {
	case ImageBitFormat_GRAY8:
		return STRIDELENGTH8(width);
	case ImageBitFormat_RGB24:
		return STRIDELENGTH24(width);
	default:
		return 0;
	}
}

ImageBase& ImageBase::operator=(const ImageBase& img) {
	if( width!=img.width || height!=img.height )
		throw 0;
	if( format==img.format)
		memcpy( data, img.data, height*getStrideSize() );
	else if( format==ImageBitFormat_RGB24 && img.format==ImageBitFormat_GRAY8 ) {
		for( uint32_t x=0; x<width; x++ )
		for( uint32_t y=0; y<height; y++ ) {
			size_t off0 = 3*x+(height-y-1)*STRIDELENGTH24(width);
			size_t off1 = 1*x+(height-y-1)*STRIDELENGTH8(width);
			data[off0+0] = img.data[off1];
			data[off0+1] = img.data[off1];
			data[off0+2] = img.data[off1];
		}
	} else
		throw 0;
	return *this;
}

bool ImageBase::save(const wchar_t* path, ImageFileFormat fmt, ImageEncoderParams* params) const {
	switch(fmt) {
	case ImageFileFormat_BMP: {
		if(format!=ImageBitFormat_RGB24)
			throw 0;
		BITMAPFILEHEADER bmfh = {0};
		BITMAPINFOHEADER bmih = {0};
		size_t size = STRIDELENGTH24(width) * height;
		size_t offset = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
		bmih.biSize				= sizeof(BITMAPINFOHEADER);
		bmih.biWidth			= width;
		bmih.biHeight			= height;
		bmih.biPlanes			= 1;
		bmih.biBitCount			= 24;
		bmih.biCompression		= 0;
		bmih.biSizeImage		= (DWORD) size;
		bmih.biXPelsPerMeter	= 0;
		bmih.biYPelsPerMeter	= 0;
		bmih.biClrUsed			= 0;
		bmih.biClrImportant		= 0;
		bmfh.bfType		= 0x4D42;
		bmfh.bfSize		= (DWORD) ( offset + size );
		bmfh.bfOffBits	= (DWORD) offset;
		HANDLE fd = CreateFileW(path,GENERIC_WRITE,0,0,OPEN_ALWAYS,0,0);
		if( fd==INVALID_HANDLE_VALUE )
			throw;
		DWORD sz;
		WriteFile(fd,&bmfh,sizeof(BITMAPFILEHEADER),&sz,0);
		WriteFile(fd,&bmih,sizeof(BITMAPINFOHEADER),&sz,0);
		WriteFile(fd,data,(DWORD)size,&sz,0);
		CloseHandle(fd);
		return false; }
	case ImageFileFormat_JPEG: {
		Gdiplus_Init();
		CLSID encoderClsid;
		if( GetEncoderClsid(L"image/jpeg", &encoderClsid) == -1 )
			return false;
		Gdiplus::PixelFormat gp_fmt;
		switch(format) {
		case ImageBitFormat_RGB24:
			gp_fmt = PixelFormat24bppRGB;
			break;
		default:
			return false;
		}
		Gdiplus::Bitmap gp_img(width,height,gp_fmt);
		Gdiplus::BitmapData gp_data;
		Gdiplus::Rect gp_rect(0,0,width,height);
		gp_img.LockBits( &gp_rect, Gdiplus::ImageLockModeWrite, gp_fmt, &gp_data );
		size_t stride = getStrideSize();
		for( uint32_t y=0; y<height; y++ )
			memcpy( (char*)gp_data.Scan0 + stride*y, data + stride*(height-y-1), stride );
		gp_img.UnlockBits(&gp_data);
		Gdiplus::EncoderParameters* gp_param = NULL;
		ULONG enc_quality;
		if( params!=NULL && ( params->mask & ImageEncoderParams_Quality ) ) {
			gp_param = (Gdiplus::EncoderParameters*) new char[sizeof(Gdiplus::EncoderParameters)];
			gp_param->Count = 1;
			gp_param->Parameter[0].Guid = Gdiplus::EncoderQuality;
			gp_param->Parameter[0].Type = Gdiplus::EncoderParameterValueTypeLong;
			gp_param->Parameter[0].NumberOfValues = 1;
			gp_param->Parameter[0].Value = &enc_quality;
			enc_quality = params->Quality;
		}
		Gdiplus::Status stat = gp_img.Save(path,&encoderClsid,gp_param);
		if( params!=NULL )
			delete[] gp_param;
		return stat == Gdiplus::Ok; }
	default:
		return false;
	}
}

ImageBase* ImageBase::load(const wchar_t* path, ImageFileFormat fmt) {
	switch(fmt) {
	case ImageFileFormat_BMP: {
		BITMAPFILEHEADER bmfh;
		BITMAPINFOHEADER bmih;
		DWORD sz;
		HANDLE fd = CreateFileW(path,GENERIC_READ,0,0,OPEN_EXISTING,0,0);
		if( fd==INVALID_HANDLE_VALUE )
			throw;
		ReadFile(fd,&bmfh,sizeof(BITMAPFILEHEADER),&sz,0);
		ReadFile(fd,&bmih,sizeof(BITMAPINFOHEADER),&sz,0);
		ImageBitFormat bfmt;
		switch(bmih.biBitCount) {
		case 8:
			bfmt = ImageBitFormat_GRAY8;
			break;
		case 24:
			bfmt = ImageBitFormat_RGB24;
			break;
		default:
			return NULL;
		}
		ImageBase* img = new ImageBase(bmih.biWidth,bmih.biHeight,bfmt);
		size_t size = img->getStrideSize() * img->height;
		SetFilePointer(fd,bmfh.bfOffBits,0,FILE_BEGIN);
		ReadFile(fd,img->data,(DWORD)size,&sz,0);
		CloseHandle(fd);
		return img; }
	case ImageFileFormat_JPEG:
	case ImageFileFormat_PNG: {
		Gdiplus_Init();
		Gdiplus::Bitmap gp_img(path);
		Gdiplus::PixelFormat gp_fmt = gp_img.GetPixelFormat();
		ImageBitFormat fmt;
		switch(gp_fmt) {
		case PixelFormat24bppRGB:
			fmt = ImageBitFormat_RGB24;
			break;
		default:
			return NULL;
		}
		ImageBase* img = new ImageBase( gp_img.GetWidth(), gp_img.GetHeight(), fmt );
		Gdiplus::BitmapData gp_data;
		Gdiplus::Rect gp_rect(0,0,img->width,img->height);
		gp_img.LockBits( &gp_rect, Gdiplus::ImageLockModeRead, gp_fmt, &gp_data );
		size_t stride = img->getStrideSize();
		for( uint32_t y=0; y<img->height; y++ )
			memcpy( img->data + stride*(img->height-y-1), (char*)gp_data.Scan0 + stride*y, stride );
		gp_img.UnlockBits(&gp_data);
		return img; }
	default:
		return NULL;
	}
}

bool ImageBase::save(const File& file, ImageFileFormat fmt) const {
	return save( String(file.path()).bytes(), fmt );
}

ImageBase* ImageBase::load(const File& file, ImageFileFormat fmt) {
	return load( String(file.path()).bytes(), fmt );
}

