/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **          Memory functions            **
 ** Version 1.1 (2016 APRIL 09)          **
 ******************************************/

#pragma once

#include <new>
#ifdef __GNUC__
#include <memory>
#endif

#include <csrutils.h>

namespace csrutils {

#ifdef _MSC_VER
#define THREAD_LOCAL __declspec(thread)
#endif
#ifdef __GNUC__
#define THREAD_LOCAL __thread
#endif

/******************************************
 **               CSR HEAP               **
 ******************************************/

struct CsrMemoryContextUser;

struct CsrMemoryContext {
	CsrMemoryContextUser* ctx; 
	void* (*alloc) (size_t, uint64_t flags);
	void (*free) (void*, uint64_t flags);
	CsrMemoryContext *prev;
};

class Heap;

const uint64_t alloc_align_mask		= 0x0F;
const uint64_t alloc_align_1byte	= 0x01;
const uint64_t alloc_align_2byte	= 0x02;
const uint64_t alloc_align_4byte	= 0x03;
const uint64_t alloc_align_8byte	= 0x04;
const uint64_t alloc_align_16byte	= 0x05;
const uint64_t alloc_align_32byte	= 0x06;

CSRLIB void heap_set_cdefault();
CSRLIB void heap_set_cppdefault();
CSRLIB void heap_set_csrthreadheap();
CSRLIB void heap_set(Heap&);
//CSRLIB void heap_push_ctx(const CsrMemoryContex& ctx);
CSRLIB void heap_push(Heap&);
CSRLIB void heap_pop();

CSRLIB void *heap_malloc(size_t size, uint64_t flags);
CSRLIB void heap_mfree(void* data, uint64_t flags);

template<typename T>
inline T* heap_alloc() {
	return (T*) heap_malloc( sizeof(T), 0 );
}

template<typename T>
inline T* heap_alloc(size_t count) {
	return (T*) heap_malloc( sizeof(T)*count, 0 );
}

inline void* heap_alloc(size_t size) {
	return heap_malloc(size,0);
}

template<typename T>
inline T* heap_alloc_ex(uint64_t flags) {
	return (T*) heap_malloc( sizeof(T), flags );
}

template<typename T>
inline T* heap_alloc_ex(size_t count, uint64_t flags) {
	return (T*) heap_malloc( sizeof(T)*count, flags );
}

inline void* heap_alloc_ex(size_t size, uint64_t flags) {
	return heap_malloc(size,flags);
}

template<typename T>
inline void heap_free(T *mem) {
	mem->~T();
	heap_mfree(mem,0);
}

template<typename T>
inline void heap_free(T *mem, size_t count) {
	for( size_t i=0; i<count; i++ )
		(mem+i)->~T();
	heap_mfree(mem,0);
}

inline void heap_free(void *mem) {
	heap_mfree(mem,0);
}

#if defined(_XMEMORY_) || defined(__GNUC__)

template<typename T>
class StdAllocator : public std::allocator<T> {
public:
    typedef size_t size_type;
    typedef T* pointer;
    typedef const T* const_pointer;

    template<typename _Tp1>
    struct rebind {
		typedef StdAllocator<_Tp1> other;
	};

    pointer allocate(size_type n, const void *hint=0) {
		return heap_alloc<T>(n);
    }

    void deallocate(pointer p, size_type n) {
		heap_free((void*)p);
    }

    StdAllocator() throw() : std::allocator<T>() { }
    StdAllocator(const StdAllocator &a) throw() : std::allocator<T>(a) { }
    template <class U>                    
    StdAllocator(const StdAllocator<U> &a) throw() : std::allocator<T>(a) { }
    ~StdAllocator() throw() { }
};

#endif



/******************************************
 **              Reference               **
 ******************************************/

struct REFERHDR {
	size_t count;
};

template<typename T>
class Reference {
	T* link;
public:
	inline Reference() : link(nullptr) { }
	inline Reference(const Reference& ref) : link(ref.link) {
		REFERHDR* hdr = (REFERHDR*)link - 1;
		hdr->count ++;
	}
	inline Reference(Reference&& ref) : link(ref.link) {
		ref.link = nullptr;
	}
	~Reference() {
		if( link!=nullptr ) {
			REFERHDR* hdr = (REFERHDR*)link - 1;
			hdr->count --;
			if( hdr->count==0 ) {
				link->~T();
				heap_free((void*)hdr);
			}
		}
	}
	inline Reference& operator= (const Reference& ref) {
		this->~Reference();
		link = ref.link;
		REFERHDR* hdr = (REFERHDR*)link - 1;
		hdr->count ++;
		return *this;
	}
	inline Reference& operator= (Reference&& ref) {
		this->~Reference();
		link = ref.link;
		ref.link = nullptr;
		return *this;
	}
	inline void alloc() {
		this->~Reference();
		REFERHDR* hdr = (REFERHDR*) heap_alloc<char>( sizeof(REFERHDR) + sizeof(T) );
		hdr->count = 1;
		link = (T*) (hdr+1);
	}
	inline T& operator*() {
		return *link;
	}
	inline operator T*() {
		return link;
	}
};


/******************************************
 **                Buffer                **
 ******************************************/

/*
	Buffer Interface
*/
class Buffer {
protected:
	void *t_data;
	size_t t_size;

public:
	Buffer() : t_data(nullptr), t_size(0) { }

	Buffer(void *data, size_t size) : t_data(data), t_size(size) { }

	inline size_t size() const {
		return t_size;
	}

	template<typename T>
	T* operator[] (T* offset) {
		return (T*) ( (size_t)t_data + (size_t)offset );
	}

	template<typename T>
	T* operator() (T* ref) {
		return (T*) ( (size_t)ref - (size_t)t_data );
	}
};



/*
	Implementation of Buffer Interface for memory buffer
*/
class BufferEx : public Buffer {
	size_t t_buffersize;

public:
	inline BufferEx() : Buffer() {
		t_data = nullptr;
		t_size = 0;
		t_buffersize = 0;
	}

	inline ~BufferEx() {
		if( t_data!=nullptr )
			heap_free(t_data);
	}

	inline BufferEx& operator= (const BufferEx& buf) {
		if( t_data!=nullptr )
			heap_free(t_data);
		t_size = buf.t_size;
		t_data = heap_alloc<char>(t_size);
		memcpy(t_data,buf.t_data,t_size);
		return *this;
	}

	inline size_t size() const {
		return t_size;
	}

	inline void expand(size_t sz) {
		if( t_buffersize >= t_size+sz )
			t_size += sz;
		else {
			if( t_buffersize==0 )
				t_buffersize = 0x500;
			while( t_buffersize < t_size+sz )
				t_buffersize *= 2;
			void *tmp = heap_alloc<char>(t_buffersize);
			if( t_size>0 && t_data!=nullptr ) {
				memcpy(tmp,t_data,t_size);
				heap_free(t_data);
			}
			t_data = tmp;
			t_size += sz;
		}
	}

	inline void setsize(size_t size) {
		void *tmp = heap_alloc<char>(size);
		if( t_size>0 && t_data!=nullptr ) {
			size_t sz = (t_size<size) ? t_size : size;
			memcpy(tmp,t_data,sz);
			heap_free(t_data);
		}
		t_data = tmp;
		t_size = size;
	}

	// Creates copy of Buffer allocated using csrutils::heap_alloc<...>
	void* getdata() {
		void *tmp = heap_alloc<char>(t_size);
		memcpy(tmp,t_data,t_size);
		return tmp;
	}
};


/******************************************
 **           Virtual Memory             **
 ******************************************/

// Memory protection

const uint64_t
	MEM_READ  = 0x001,
	MEM_WRITE = 0x002,
	MEM_EXEC  = 0x004;

// Memory allocation flags

const uint64_t
	MEM_SHARED	= 0x001,	// Creates a shared memory object (a map)
	MEM_LARGE	= 0x002;	// Allocates large (huge) pages


struct MemorySection_Ctx;

class MemorySection {
	MemorySection_Ctx *ctx;
	void *t_data;
	size_t t_size;

public:
	inline MemorySection() : ctx(nullptr), t_data(nullptr), t_size(0) { }
	MemorySection(size_t size, uint64_t flags, uint64_t prot);
	MemorySection(const MemorySection&);
	MemorySection(MemorySection&&);
	virtual ~MemorySection();
	MemorySection& operator= (const MemorySection&);
	MemorySection& operator= (MemorySection&&);

	inline void* operator* () const { return t_data; }
	inline size_t size() const { return t_size; }

	static size_t getPageSize();
	static size_t getLargePageSize();
};

};
