/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         Multithreading utils         **
 ** Version: 0.2 (2016 APRIL 16)         **
 ******************************************/

#pragma once

#include "csrutils.h"

#define __thread __declspec(thread)

namespace csrutils {

/**************************************
 **            SyncEvent             **
 **************************************/

#ifdef _WINDOWS_
typedef HANDLE CSR_HEVENT;
#else
struct CSRSYNCEVENT;
typedef CSRSYNCEVENT* CSR_HEVENT;
#endif

class SyncEvent {
	CSR_HEVENT nevent;

public:
	SyncEvent(bool manualReset, bool initState);
	~SyncEvent();

	void set();
	void reset();
	void wait();
	void wait(uint32_t time);
	static void waitAll(SyncEvent* events, size_t count);
	static void waitAll(SyncEvent* events, size_t count, uint32_t time);
	static void waitSingle(SyncEvent* events, size_t count);
	static void waitSingle(SyncEvent* events, size_t count, uint32_t time);

	SyncEvent(const SyncEvent&) = delete;
	void operator=(const SyncEvent&) = delete;

};

typedef SyncEvent* PSyncEvent;



/**************************************
 **           SyncObject             **
 **             (Mutex)              **
 **************************************/

struct SyncObject_Ctx;

class SyncObject {
	SyncObject_Ctx* ctx;

public:
	SyncObject();
	~SyncObject();

	void uses_inc();
	// Decrement number of uses. Returns true if no more uses.
	bool uses_dec();
	size_t uses();

	void lock();
	void unlock();
};



/**************************************
 **             Thread               **
 **************************************/

#ifdef _WINDOWS_
typedef HANDLE CSR_HTHREAD;
#else
typedef void* CSR_HTHREAD;
#endif

class ThreadJob {
public:
	virtual void run() = 0;
};

class ThreadBase;

class Thread {
	ThreadBase* base;

public:
	Thread();
	Thread(ThreadJob* runjob);
	Thread(ThreadBase* base);
	Thread(const Thread&);
	~Thread();

	void start();
	void join();

	Thread& operator= (const Thread&);
	bool operator==(const Thread&) const;
	inline bool operator!=(const Thread& t) const { return ! (*this==t); }

	static void sleep(int time);
	static Thread current();
};



/**************************************
 **             Job Queue            **
 **************************************/

/*
struct JobQueue_Task;

typedef void (*JobQueue_Proc) (JobQueue_Task *task);

struct JobQueue_Task {
	void *data;		// allocated with 'Objmon_alloc'
	JobQueue_Proc proc;	// doesn't delete 'data'
};

struct JobQueue_Ctx;

class JobQueue {
	JobQueue_Ctx *ctx;

public:
	JobQueue(int threads);
	void push(JobQueue_Task task);
	bool pop(JobQueue_Task &task);
	void clear();
};

*/


/**************************************
 **            Sync Queue            **
 **************************************/

struct SyncQueueBase_Ctx;

class SyncQueueBase {
	SyncQueueBase_Ctx* ctx;
	size_t maxcount;
	size_t p_front;
	size_t p_back;
	size_t p_count;
	void** p_data;

public:
	SyncQueueBase(size_t maxcount);
	~SyncQueueBase();

	void push(void*);
	void* pop();
};

template<typename T> class SyncQueue {
	SyncQueueBase base;
	int maxcount;

public:
	SyncQueue(int _maxcount) : maxcount(_maxcount), base(_maxcount) {
	}

	void push(T* object) {
		base.push(object);
	}

	T* pop() {
		return (T*) base.pop();
	}
};


/**************************************
 **        Execution Monitor         **
 **************************************/

struct ExecitionMonitor_Ctx;

class ExecitionMonitor {
	ExecitionMonitor_Ctx* ctx;

public:
	ExecitionMonitor();
	~ExecitionMonitor();

	// Methods called by controller
	void abort();
	void pause();
	void unpause();

	// Methods called by execution procedure threads
	void proc_attend(bool& aborted);
	void proc_enter();
	void proc_exit();
};


/**************************************
 **          Thread Monitor          **
 **************************************/

class ThreadMonitor;
class ThreadMonitorChild;
struct ThreadMonitor_Ctx;

#ifdef _CSRUTILS_INT_
struct ThreadMonitor_Job;
#endif

typedef bool (*ThreadMonitor_Proc)(ThreadMonitor& monitor);

class ThreadMonitor {
	ThreadMonitor_Ctx* ctx;
	void* pUserData;

	bool checkmainthread();
#ifdef _CSRUTILS_INT_
	ThreadMonitor_Job& getcurrentjob();
#endif

public:
	ThreadMonitor();
	~ThreadMonitor();

	ThreadMonitorChild createchild(ThreadMonitor_Proc proc, double weight, void* pUser);
	void waitforchilds();
	void abortchilds();
	void setmainthread();

	void pushjob(double weight);
	void popjob();
	void setjob(double per);
	bool checkabort();
	void setcompleted();

	double getpers();

	inline void setuserdata(void* pUser) { pUserData = pUser; }
	void* getuserdata() { return pUserData; }

	void cmd_abort();
	void cmd_pause();
	void cmd_unpause();
};

class ThreadMonitorChild {
	ThreadMonitor& monitor;
	ThreadMonitor_Proc proc;

	ThreadMonitorChild(ThreadMonitor& monitor, ThreadMonitor_Proc proc);

public:
	void run();
	void pause();
	void abort();

	friend class ThreadMonitor;
};


/**************************************
 **          Execution Pool          **
 **************************************/

class ExecutionPool {
public:
	struct Context;
	Context *ctx;

public:

	class Item {
	public:
		virtual ~Item() { }
		virtual void execute() = 0;
	};

	class Task {
		struct Context;
		Context *ctx;

	public:
		Task();
		virtual ~Task();

		void setPool(ExecutionPool* pool);
		bool pushItem(Item* item);
		Item* popItem();
		void doneItem();
		void wait();
		void threadEnter();
		void threadLeave();
		
		virtual void handleBegin() { }
		virtual void handleCompletion() { }
		virtual void handleAbort() { }
		virtual void handleCancel() { }
		virtual void handleThreadEnter() { }
		virtual void handleThreadLeave() { }
	};

	ExecutionPool(size_t nThreads);
	~ExecutionPool();

	bool addTask(Task *task);
	bool removeTask(Task *task);
	Task *getTask();
	void wait();
};

};
