/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         Multithreading utils         **
 ** ExecutionPool implementation         **
 ** Version: 0.2 (2016 APRIL 16)         **
 ******************************************/

#include <csrutils.h>
#include <csrmthread.h>
#include <csrmthread_int.h>
#include <csrmem.h>
#include <list>

using namespace csrutils;

class ExecutionPool_Job : public ThreadJob {
public:
	ExecutionPool* pool;
	ExecutionPool::Context* pool_ctx;
	void run();
};

struct ExecutionPool::Context {
	size_t nThreads;
	Thread* threads;
	ExecutionPool_Job* job;
	bool f_destroy;
	thread_event_t s_exited;
	volatile size_t nThreadsRunning;
	std::list<Task*> tasks;
	thread_mutex_t mutex;
	thread_cond_t signal_notempty;
};

struct ExecutionPool::Task::Context {
	ExecutionPool* pool;
	std::list<Item*> items;
	thread_mutex_t mutex;
	thread_cond_t signal_notempty;
	bool f_completed;
	volatile size_t nTasksRunning;
	volatile size_t nThreadsRunning;
};

ExecutionPool::ExecutionPool(size_t nThreads) {
	ctx = new( heap_alloc<Context>() ) Context();
	ctx->nThreads = nThreads;
	ctx->threads  = heap_alloc<Thread>(nThreads);
	ctx->job      = new( heap_alloc<ExecutionPool_Job>() ) ExecutionPool_Job();
	ctx->job->pool       = this;
	ctx->job->pool_ctx   = ctx;
	ctx->f_destroy       = false;
	ctx->nThreadsRunning = nThreads;
	thread_event_init(&ctx->s_exited,true,false);
	thread_mutex_init(&ctx->mutex);
	thread_cond_init(&ctx->signal_notempty);
	for( size_t i=0; i<nThreads; i++ )
		new(&ctx->threads[i]) Thread(ctx->job);
	for( size_t i=0; i<nThreads; i++ )
		ctx->threads[i].start();
	Thread::sleep(100);
}

ExecutionPool::~ExecutionPool() {
	ctx->f_destroy = true;
	thread_cond_signal(&ctx->signal_notempty);
	wait();
	heap_free(ctx->job);
	heap_free(ctx->threads,ctx->nThreads);
	heap_free(ctx);
}
bool ExecutionPool::addTask(Task *task) {
	task->setPool(this);
	thread_mutex_lock(&ctx->mutex);
	bool f_signal = ctx->tasks.size()==0;
	ctx->tasks.push_back(task);
	thread_mutex_unlock(&ctx->mutex);
	if( f_signal )
		thread_cond_signal(&ctx->signal_notempty);
	return true;
}

bool ExecutionPool::removeTask(Task *task) {
	thread_mutex_lock(&ctx->mutex);
	for( auto iter=ctx->tasks.begin(); iter!=ctx->tasks.end(); iter++ ) {
		if( *iter == task ) {
			ctx->tasks.erase(iter);
			thread_mutex_unlock(&ctx->mutex);
			return true;
		}
	}
	thread_mutex_unlock(&ctx->mutex);
	return false;
}

ExecutionPool::Task *ExecutionPool::getTask() {
	thread_mutex_lock(&ctx->mutex);
	while( ctx->tasks.size()==0 && (!ctx->f_destroy) )
		thread_cond_wait(&ctx->signal_notempty,&ctx->mutex);
	if( ctx->f_destroy ) {
		thread_mutex_unlock(&ctx->mutex);
		return nullptr;
	}
	Task *task = ctx->tasks.front();
	thread_mutex_unlock(&ctx->mutex);
	return task;
}

void ExecutionPool::wait() {
	while( ctx->nThreadsRunning > 0 )
		thread_event_wait(&ctx->s_exited);
}

void ExecutionPool_Job::run() {
	//interlocked_increment(&pool_ctx->nThreadsRunning);
	for(;;) {
		ExecutionPool::Task* task = pool->getTask();
		if( task == nullptr )
			break;
		task->threadEnter();
		for(;;) {
			ExecutionPool::Item* item = task->popItem();
			if( item == nullptr )
				break;
			item->execute();
			task->doneItem();
			heap_free(item);
		}
		task->threadLeave();
	}
	interlocked_decrement(&pool_ctx->nThreadsRunning);
	if( pool_ctx->nThreadsRunning == 0 )
		thread_event_set(&pool_ctx->s_exited);
}

ExecutionPool::Task::Task() {
	ctx = new( heap_alloc<Context>() ) Context();
	ctx->f_completed     = false;
	ctx->nTasksRunning   = 0;
	ctx->nThreadsRunning = 0;
	thread_mutex_init(&ctx->mutex);
	thread_cond_init(&ctx->signal_notempty);
}

ExecutionPool::Task::~Task() {
	thread_mutex_deinit(&ctx->mutex);
	heap_free(ctx);
}

void ExecutionPool::Task::setPool(ExecutionPool* pool) {
	ctx->pool = pool;
}

bool ExecutionPool::Task::pushItem(Item* item) {
	thread_mutex_lock(&ctx->mutex);
	if( ctx->items.size()==0 )
		thread_cond_signal(&ctx->signal_notempty);
	ctx->items.push_back(item);
	thread_mutex_unlock(&ctx->mutex);
	return true;
}

ExecutionPool::Item* ExecutionPool::Task::popItem() {
	thread_mutex_lock(&ctx->mutex);
	while( ctx->items.size()==0 && (!ctx->f_completed) )
		thread_cond_wait(&ctx->signal_notempty,&ctx->mutex);
	if( ctx->f_completed ) {
		thread_mutex_unlock(&ctx->mutex);
		return nullptr;
	}
	Item *item = ctx->items.front();
	ctx->items.pop_front();
	interlocked_increment(&ctx->nTasksRunning);
	thread_mutex_unlock(&ctx->mutex);
	return item;
}

void ExecutionPool::Task::doneItem() {
	thread_mutex_lock(&ctx->mutex);
	interlocked_decrement(&ctx->nTasksRunning);
	if( ctx->nTasksRunning == 0 && ctx->items.size()==0 ) {
		ctx->f_completed = true;
		thread_cond_signal(&ctx->signal_notempty);
		ctx->pool->removeTask(this);
	}
	thread_mutex_unlock(&ctx->mutex);
}

void ExecutionPool::Task::wait() {
	thread_mutex_lock(&ctx->mutex);
	while( ! ctx->f_completed )
		thread_cond_wait(&ctx->signal_notempty,&ctx->mutex);
	thread_mutex_unlock(&ctx->mutex);
}

void ExecutionPool::Task::threadEnter() {
	thread_mutex_lock(&ctx->mutex);
	if( ctx->nThreadsRunning == 0 )
		handleBegin();
	interlocked_increment(&ctx->nThreadsRunning);
	thread_mutex_unlock(&ctx->mutex);
	handleThreadEnter();
}

void ExecutionPool::Task::threadLeave() {
	handleThreadLeave();
	thread_mutex_lock(&ctx->mutex);
	interlocked_decrement(&ctx->nThreadsRunning);
	if( ctx->nThreadsRunning == 0 )
		handleCompletion();
	thread_mutex_unlock(&ctx->mutex);
}
