/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         IO stream utilities          **
 ** Version: 1.0 (2016 MARCH 27)         **
 ******************************************/

#pragma once

#include <exception>

#include "csrutils.h"

namespace csrutils {

/*
Codes returned by read and write functions.
Code IO_OK = 0 indicates that operation succeeded,
and other codes indicates some errors.
Bigger numbers can be used to define other (specific) error codes.
*/
const int
	IO_OK = 0,			// Operation succeded, no error
	IO_ERROR = 1,		// Unknown error
	IO_EOF = 2,			// End of stream
	IO_NOTFOUND = 3;	// Resource (file, etc.) not found


class IOException : public std::exception {
	const int code;
public:
	inline IOException(int code) : code(code) { }
	const char* what() const throw() {
		return "Unkown IO error";
	}
};

class IOStream {
public:
	virtual ~IOStream() { }

/*
Read 'size' bytes from input stream into 'data' buffer.
Number of read bytes stored int 'read' variable.
In case of an error returns non-zero error and 'read'
is number of successfuly read bytes before error eccured.
Error code doesn't mean that no data has been read, 'read'=0 does.
Condition 'read' < 'size' doesn't indicate any error.
End of input stream indicated by returning IO_EOF.
*/
	virtual int read(void *data, size_t size, size_t& read) = 0;

/*
Write 'size' bytes into output stream from 'data' buffer.
Number of written bytes stored int 'written' variable.
In case of an error returns non-zero error and 'written'
is number of successfuly written bytes before error eccured.
Error code doesn't mean that no data has been written, 'written'=0 does.
Condition 'written' < 'size' doesn't indicate any error.
*/
	virtual int write(const void *data, size_t size, size_t& written) = 0;

/*
Read data from input stream into output stream until error (e.g. EOF).
Data has been read by chunks (e.g. 512 bytes) and written.
*/
	inline static int transfer(IOStream *input, IOStream *output) {
		size_t buf_size = 512;
		uint8_t* buffer = new uint8_t[buf_size];
		size_t sz;
		int res;
		do {
			res = input->read(buffer,buf_size,sz);
			if( res!=0 || sz==0 )
				break;
			res = output->write(buffer,sz,sz);
		} while( res==0 && sz>0 );
		return res;
	}
};

/*struct FileReadStreamCtx;
class FileReadStream : public IOStream {
	FileReadStreamCtx *ctx;

public:
	FileReadStream(const char *path);
	~FileReadStream();

	int read(void *data, size_t size, size_t *readed);
	int write(const void *data, size_t size, size_t *written);

	uint64_t getsize();
	uint64_t getposition();
};*/

class MemoryStream : public IOStream {
	uint8_t* c_data;
	size_t c_size;
	size_t c_offset;

public:
	inline MemoryStream(void *data, size_t size) {
		c_data   = (uint8_t*) data;
		c_size   = size;
		c_offset = 0;
	}

	inline ~MemoryStream() { }

	int read(void *data, size_t size, size_t& read) {
		size_t sz = (size>c_size-c_offset)?(c_size-c_offset):size;
		read = sz;
		memcpy(data,c_data+c_offset,sz);
		c_offset += sz;
		return 0;
	}

	int write(const void *data, size_t size, size_t& written) {
		size_t sz = (size>c_size-c_offset) ? (c_size-c_offset) : size;
		written = sz;
		memcpy(c_data+c_offset,data,sz);
		c_offset += sz;
		return 0;
	}

	// Set end of stream to current position
	inline void cut() {
		c_size = c_offset;
	}

	// Reset offset into buffer begining
	inline void restart() {
		c_offset = 0;
	}

	inline size_t offset() {
		return c_offset;
	}
};

class FileStream : public IOStream {
public:
/*
Read 'size' bytes from input stream into 'data' buffer.
Number of read bytes stored int 'read' variable.
In case of an error returns non-zero error and 'read'
is number of successfuly read bytes before error eccured.
Error code doesn't mean that no data has been read, 'read'=0 does.
Condition 'read' < 'size' doesn't indicate any error.
End of input stream indicated by returning IO_EOF.
*/
	virtual int read(void *data, size_t size, size_t& read) = 0;

/*
Write 'size' bytes into output stream from 'data' buffer.
Number of written bytes stored int 'written' variable.
In case of an error returns non-zero error and 'written'
is number of successfuly written bytes before error eccured.
Error code doesn't mean that no data has been written, 'written'=0 does.
Condition 'written' < 'size' doesn't indicate any error.
*/
	virtual int write(const void *data, size_t size, size_t& written) = 0;

/*
Get size of file. In case of an error -1 is returned.
*/
	virtual uint64_t getsize() = 0;

/*
Get offset of file pointer relative to file begining.
To be used with setposition() function.
In case of an error -1 is returned.
*/
	virtual uint64_t getposition() = 0;

/*
Set offset of file pointer relative to file begining.
To be used with getposition() function.
In case of an error -1 is returned.
*/
	virtual bool setposition(uint64_t pos) = 0;

/*
Truncate file to current pointer position.
*/
	virtual bool setend() = 0;

	//FilePointer pos() { return FilePointer(*this); }
};

/*
BufferStream is a variant of FileStream which stores all the date
in the memory and provides file-like interface to access it, i.e.
virtual (in-memory) file.
Also it defines additional method getdata() which returns pointer
to raw data representation in the memory.
*/

class BufferStream : public FileStream {
	char *s_data;
	size_t s_bufsize;
	size_t s_size;
	size_t s_offset;

public:
	BufferStream();
	~BufferStream();

	int read(void *data, size_t size, size_t& read);
	int write(const void *data, size_t size, size_t& written);
	uint64_t getsize();
	uint64_t getposition();
	bool setposition(uint64_t pos);
	bool setend();
	void *getdata();
};

struct NestedStopReader_Ctx;

class NestedStopReader : public FileStream {
	NestedStopReader_Ctx *ctx;
	FileStream *stream;

public:
	NestedStopReader(FileStream *stream);
	~NestedStopReader();
	int read(void *data, size_t size, size_t& read);
	int write(const void *data, size_t size, size_t& written);
	uint64_t getsize();
	bool setsize(uint64_t size);
	uint64_t getposition();
	bool setposition(uint64_t pos);
	bool setend();

	void setstop(const void *data, size_t size);
	void unsetstop(const void *data, size_t size);
	void setstop(const char *data);
	void unsetstop(const char *data);
};

};