/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **          String functions            **
 **         [Additional header]          **
 ** Version 1.1 (2016 MARCH 27)          **
 ******************************************
 ** WARNING: Do not include this file,   **
 ** it is included via csrstr.h          **
 ******************************************/

#pragma once

//#include <tchar.h>

namespace csrutils {

/*
8-bit ANSI String Interface without Buffer
Char type: char
Bits: 8
*/	
class StringA {
	StringEncoding tenc;
	char *tdata;
	size_t tlen;

public:
	inline StringA();
	inline ~StringA();
	inline StringA(const StringA&);
	inline StringA(const char*);
	inline StringA(const char*, StringEncoding);

	inline StringA& operator= (const StringA&);
	inline StringAB operator+ (const StringA&) const;

	inline bool operator== (const StringA& s) const;
	inline bool operator== (const char*) const;
	inline bool operator!= (const StringA&) const;
	inline bool operator!= (const char*) const;
	inline bool operator< (const StringA& s) const;
	inline bool operator> (const StringA& s) const;
	inline bool operator<= (const StringA&) const;
	inline bool operator>= (const StringA&) const;
	inline char operator[] (size_t pos) const { return tdata[pos]; }
	
	/* Retuns length of string in chars */
	inline size_t len() const { return tlen; }

	/* Return size of string buffer in bytes */
	inline size_t size() const { return sizeof(char)*tlen; }

	// returns pointer to chars
	// might be invalidated by further calls
	inline const char* bytes() const { return tdata; }

	// returns pointer to chars
	// might be invalidated by further calls
	inline char* bytes() { return tdata; }

	inline StringEncoding getencoding() const { return tenc; }
	inline void setencoding(StringEncoding enc) { tenc = enc; }

	/*
	Copies chars to buffer (must be preallocated)
	Recommended allocation as heap_alloc<char>( str.len()+1 )
	*/
	inline void copy(char *buffer) const;

	inline size_t strpos(char c) const;
	inline size_t strrpos(char c) const;

	inline StringAB substr(size_t begin) const;
	inline StringAB substr(size_t begin, size_t length) const;
	StringAB escapes(StringEscapeFlags flags);
	StringAB invescapes(StringEscapeFlags flags);

	//StringB convert(StringEncoding encoding) const;

	friend class StringAB;
};


/*
8-bit ANSI String Interface with Buffer
Char type: char
Bits: 8
*/	
class StringAB : public StringA {
public:
	inline StringAB() { };
	inline ~StringAB();
	inline StringAB(const StringA&);
	inline StringAB(const StringAB& s) : StringAB( (StringA&)s ) { }
	inline StringAB(StringA&&);
	inline StringAB(const char*);
	inline StringAB(const char*, StringEncoding);
	inline StringAB(const char*, size_t);
	inline StringAB(const char*, size_t, StringEncoding);
	StringAB(const String&, StringEncoding);

	inline StringA& operator=(const StringA& s);
	inline StringA& operator=(const StringAB& s) { return (*this) = (StringA&) s; }
	inline StringA& operator=(StringAB&& s);
	StringA& operator= (const String&);
	inline void operator+= (const StringA& s);
};

inline StringA::StringA() {
	tdata = nullptr;
	tlen  = 0;
	tenc  = get_default_encoding();
}

inline StringA::~StringA() { }

inline StringAB::~StringAB() {
	if( tdata!=nullptr )
		heap_free(tdata);
	tdata = nullptr;
}

inline StringA::StringA(const StringA& s) {
	tdata = s.tdata;
	tlen  = s.tlen;
	tenc  = s.tenc;
}

inline StringAB::StringAB(const StringA& s) : StringA() {
	if( s.tdata != nullptr ) {
		tdata = heap_alloc<char>(s.tlen+1);
		tlen  = s.tlen;
		tenc  = s.tenc;
		copy_array(s.tdata,s.tdata+(tlen+1),tdata);
	} else {
		tdata = nullptr;
		tlen  = 0;
		tenc  = s.tenc;
	}
}

inline StringAB::StringAB(StringA&& s) {
	tdata   = s.tdata;
	tlen    = s.tlen;
	tenc    = s.tenc;
	s.tdata = nullptr;
	s.tlen  = 0;
}

inline StringA::StringA(const char* data) {
	tdata = (char*) data;
	tlen  = strlen(data);
	tenc  = get_default_encoding();
}

inline StringA::StringA(const char* data, StringEncoding enc) {
	tdata = (char*) data;
	tlen  = strlen(data);
	tenc  = enc;
}

inline StringAB::StringAB(const char* data) {
	tlen  = strlen(data);
	tdata = heap_alloc<char>(tlen+1);
	tenc  = get_default_encoding();
	copy_array(data,data+(tlen+1),tdata);
}

inline StringAB::StringAB(const char* data, StringEncoding enc) {
	tlen  = strlen(data);
	tdata = heap_alloc<char>(tlen+1);
	tenc  = enc;
	copy_array(data,data+(tlen+1),tdata);
}

inline StringAB::StringAB(const char* data, size_t len) {
	tlen  = len;
	tdata = heap_alloc<char>(tlen+1);
	tenc  = get_default_encoding();
	copy_array(data,data+tlen,tdata);
	tdata[tlen] = '\0';
}

inline StringAB::StringAB(const char* data, size_t len, StringEncoding enc) {
	tlen  = len;
	tdata = heap_alloc<char>(tlen+1);
	tenc  = enc;
	copy_array(data,data+tlen,tdata);
	tdata[tlen] = '\0';
}

inline StringA& StringA::operator=(const StringA& s) {
	tdata = s.tdata;
	tlen  = s.tlen;
	return *this;
}

inline StringA& StringAB::operator=(const StringA& s) {
	if( tdata != nullptr )
		heap_free(tdata);
	if( s.tdata != nullptr ) {
		tlen  = s.tlen;
		tdata = heap_alloc<char>(tlen+1);
		tenc  = s.tenc;
		copy_array( s.tdata, s.tdata+(tlen+1), tdata );
	} else {
		tdata = nullptr;
		tlen  = 0;
		tenc  = s.tenc;
	}
	return *this;
}

inline StringA& StringAB::operator=(StringAB&& s) {
	if( tdata!=NULL )
		heap_free(tdata);
	tlen  = s.tlen;
	tdata = s.tdata;
	tenc  = s.tenc;
	s.tlen  = 0;
	s.tdata = NULL;
	return *this;
}

inline StringAB StringA::operator+ (const StringA& s) const {
	StringAB res;
	res.tdata = heap_alloc<char>(tlen+s.tlen+1);
	res.tlen  = tlen+s.tlen;
	res.tenc  = tenc;
	copy_array( tdata, tdata+(tlen+1), res.tdata );
	copy_array( s.tdata, s.tdata+(s.tlen+1), res.tdata+tlen );
	return res;
}

inline void StringAB::operator+= (const StringA& s)  {
	*this = *this + s;
}

inline bool StringA::operator==(const StringA& s) const {
	if( s.tlen!=tlen )
		return false;
	for( size_t i=0; i<tlen; i++ )
		if( tdata[i] != s.tdata[i] )
			return false;
	return true;
}

inline bool StringA::operator==(const char* s) const {
	for( size_t i=0; ; i++ ) {
		if( tlen==i && s[i]=='\0' )
			return true;
		if( tdata[i] != s[i] )
			return false;
	}
}

inline bool StringA::operator!=(const StringA& s) const {
	return !( (*this) == s );
}

inline bool StringA::operator!=(const char* s) const {
	return !( (*this) == s );
}

inline bool StringA::operator< (const StringA& s) const {
	for( size_t i=0; ; i++ ) {
		if( tdata[i] < s.tdata[i] )
			return true;
		if( tdata[i] > s.tdata[i] )
			return false;
		if( tdata[i]=='\0' && s.tdata[i]=='\0' )
			return false;
	}
}

inline bool StringA::operator> (const StringA& s) const {
	for( size_t i=0; ; i++ ) {
		if( tdata[i] > s.tdata[i] )
			return true;
		if( tdata[i] < s.tdata[i] )
			return false;
		if( tdata[i]=='\0' && s.tdata[i]=='\0' )
			return false;
	}
}

inline bool StringA::operator<= (const StringA& s) const {
	return !( *this > s );
}

inline bool StringA::operator>= (const StringA& s) const {
	return !( *this < s );
}

inline void StringA::copy(char *buffer) const {
	copy_array(tdata,tdata+(tlen+1),buffer);
}


inline size_t StringA::strpos(char c) const {
	for( size_t i=0; i<tlen; i++ )
		if( tdata[i]==c )
			return i;
	return -1;
}

inline size_t StringA::strrpos(char c) const {
	for( size_t i=tlen-1; i>=0; i-- )
		if( tdata[i]==c )
			return i;
	return -1;
}

inline StringAB StringA::substr(size_t begin) const {
	if( begin<0 || begin>tlen )
		throw 0;
	return StringAB(tdata+begin,tlen-begin);
}

inline StringAB StringA::substr(size_t begin, size_t length) const {
	if( begin<0 || begin+length>tlen )
		throw 0;
	return StringAB(tdata+begin,length);
}

}
