/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **          String functions            **
 **         [Additional header]          **
 ** Version 1.0 (2015 DECEMBER 04)       **
 ******************************************
 ** WARNING: Do not include this file,   **
 ** it is included via csrstr.h          **
 ******************************************/

#pragma once

//#include <tchar.h>
#include <string.h>

namespace csrutils {

class StringU {
protected:
	char *tdata;
	size_t tlen;

public:
	inline StringU();
	inline ~StringU();
	inline StringU(const StringU&);
	inline StringU(StringU&&);
	inline StringU(const char* data);
	inline StringU(const char* data, size_t len);

	inline StringU& operator=(const StringU& s);
	inline StringU& operator=(StringU&& s);

	inline StringU operator+ (const StringU& s) const;
	inline void operator+= (const StringU& s);

	inline bool operator==(const StringU& s) const;
	inline bool operator==(const char* s) const;
	inline bool operator!=(const StringU& s) const;
	inline bool operator!=(const char* s) const;
	inline bool operator< (const StringU& s) const;
	inline bool operator> (const StringU& s) const;
	inline bool operator<= (const StringU& s) const;
	inline bool operator>= (const StringU& s) const;

	inline size_t len() const { return tlen; }
	inline size_t size() const { return sizeof(char)*tlen; }

	// returns pointer to bytes
	// might be invalidated by further calls
	inline const char* bytes() const { return tdata; }
	// returns pointer to chars
	// might be invalidated by further calls
	inline char* bytes() { return tdata; }
	// copies chars into buffer (must be allocated)
	inline void copy(char *buffer) const;
	inline wchar_t operator[] (size_t pos) const;

	inline size_t strpos(char c) const;
	inline size_t strrpos(char c) const;

	inline StringU substr(size_t begin) const;
	inline StringU substr(size_t begin, size_t length) const;
	//StringU escapes(StringEscapeFlags flags);
	//StringU invescapes(StringEscapeFlags flags);
};

inline StringU::StringU() {
	tdata = NULL;
	tlen  = 0;
}

inline StringU::~StringU() {
	if( tdata!=NULL )
		heap_free(tdata);
}

inline StringU::StringU(const StringU& s) {
	if( s.tdata!=NULL ) {
		tdata = heap_alloc<char>(s.tlen+1);
		tlen = s.tlen;
		copy_array(s.tdata,s.tdata+(tlen+1),tdata);
	} else {
		tdata = NULL;
		tlen  = 0;
	}
}

inline StringU::StringU(StringU&& s) {
	tdata = s.tdata;
	tlen  = s.tlen;
	s.tdata = NULL;
	s.tlen  = 0;
}

inline StringU::StringU(const char* data) {
	tlen  = strlen(data);
	tdata = heap_alloc<char>(tlen+1);
	copy_array(data,data+(tlen+1),tdata);
}

inline StringU::StringU(const char* data, size_t len) {
	tlen  = len;
	tdata = heap_alloc<char>(tlen+1);
	copy_array(data,data+(tlen+1),tdata);
	tdata[tlen] = '\0';
}

inline StringU& StringU::operator=(const StringU& s) {
	if( tdata!=NULL )
		heap_free(tdata);
	if( s.tdata != NULL ) {
		tlen  = s.tlen;
		tdata = heap_alloc<char>(tlen+1);
		copy_array( s.tdata, s.tdata+(tlen+1), tdata );
	} else {
		tdata = NULL;
		tlen  = 0;
	}
	return *this;
}

inline StringU& StringU::operator=(StringU&& s) {
	if( tdata!=NULL )
		heap_free(tdata);
	tlen  = s.tlen;
	tdata = s.tdata;
	s.tlen  = 0;
	s.tdata = NULL;
	return *this;
}

inline StringU StringU::operator+ (const StringU& s) const {
	StringU res;
	res.tdata = heap_alloc<char>(tlen+s.tlen+1);
	res.tlen  = tlen+s.tlen;
	copy_array( tdata, tdata+(tlen+1), res.tdata );
	copy_array( s.tdata, s.tdata+(tlen+1), res.tdata+tlen );
	return res;
}

inline void StringU::operator+= (const StringU& s)  {
	*this = *this + s;
}

inline bool StringU::operator==(const StringU& s) const {
	if( s.tlen!=tlen )
		return false;
	for( size_t i=0; i<tlen; i++ )
		if( tdata[i] != s.tdata[i] )
			return false;
	return true;
}

inline bool StringU::operator==(const char* s) const {
	for( size_t i=0; ; i++ ) {
		if( tlen==i && s[i]=='\0' )
			return true;
		if( tdata[i] != s[i] )
			return false;
	}
}

inline bool StringU::operator!=(const StringU& s) const {
	return !( (*this) == s );
}

inline bool StringU::operator!=(const char* s) const {
	return !( (*this) == s );
}

inline bool StringU::operator< (const StringU& s) const {
	for( size_t i=0; ; i++ ) {
		if( tdata[i] < s.tdata[i] )
			return true;
		if( tdata[i] > s.tdata[i] )
			return false;
		if( tdata[i]=='\0' && s.tdata[i]=='\0' )
			return false;
	}
}

inline bool StringU::operator> (const StringU& s) const {
	for( size_t i=0; ; i++ ) {
		if( tdata[i] > s.tdata[i] )
			return true;
		if( tdata[i] < s.tdata[i] )
			return false;
		if( tdata[i]=='\0' && s.tdata[i]=='\0' )
			return false;
	}
}

inline bool StringU::operator<= (const StringU& s) const {
	return !( *this > s );
}

inline bool StringU::operator>= (const StringU& s) const {
	return !( *this < s );
}

inline void StringU::copy(char *buffer) const {
	copy_array(tdata,tdata+(tlen+1),buffer);
}


inline size_t StringU::strpos(char c) const {
	for( size_t i=0; i<tlen; i++ )
		if( tdata[i]==c )
			return i;
	return -1;
}

inline size_t StringU::strrpos(char c) const {
	for( size_t i=tlen-1; i>=0; i-- )
		if( tdata[i]==c )
			return i;
	return -1;
}

inline StringU StringU::substr(size_t begin) const {
	if( begin<0 || begin>tlen )
		throw 0;
	return StringU(tdata+begin,tlen-begin);
}

inline StringU StringU::substr(size_t begin, size_t length) const {
	if( begin<0 || begin+length>tlen )
		throw 0;
	return StringU(tdata+begin,length);
}


}
