/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **          String functions            **
 **         [Additional header]          **
 ** Version 1.1 (2016 MARCH 27)          **
 ******************************************
 ** WARNING: Do not include this file,   **
 ** it is included via csrstr.h          **
 ******************************************/

#pragma once

#include <wchar.h>

namespace csrutils {

/*
16-bit Unicode String Interface without Buffer
Char type: wchar_t
Bits: 16 (or 32 on linux)
*/
class String {
protected:
	wchar_t *tdata;
	size_t tlen;

public:
	inline String();
	inline ~String();
	inline String(const String&);
	inline String(const wchar_t*);

	inline String& operator= (const String&);
	inline StringB operator+ (const String&) const;

	inline bool operator== (const String&) const;
	inline bool operator== (const wchar_t*) const;
	inline bool operator!= (const String&) const;
	inline bool operator!= (const wchar_t*) const;
	inline bool operator< (const String&) const;
	inline bool operator> (const String&) const;
	inline bool operator<= (const String&) const;
	inline bool operator>= (const String&) const;
	inline wchar_t operator[] (size_t pos) const { return tdata[pos]; }

	/* Retuns length of string in chars */
	inline size_t len() const { return tlen; }

	/* Return size of string buffer in bytes */
	inline size_t size() const { return sizeof(wchar_t)*tlen; }

	// returns pointer to chars
	// might be invalidated by further calls
	inline const wchar_t* bytes() const { return tdata; }

	// returns pointer to chars
	// might be invalidated by further calls
	inline wchar_t* bytes() { return tdata; }

	/*
	Copies chars to buffer (must be preallocated)
	Recommended allocation as heap_alloc<wchar_t>( str.len()+1 )
	*/
	inline void copy(wchar_t *buffer) const;

	inline size_t strpos(wchar_t c) const;
	inline size_t strrpos(wchar_t c) const;

	inline StringB substr(size_t begin) const;
	inline StringB substr(size_t begin, size_t length) const;
	StringB escapes(StringEscapeFlags flags);
	StringB invescapes(StringEscapeFlags flags);

	//StringA convert(StringEncoding encoding) const;
	//StringA utf8() const;

	friend class StringB;
};


/*
16-bit Unicode String Interface with Buffer
Char type: wchar_t
Bits: 16 (or 32 on linux)
*/
class StringB : public String {
public:
	inline StringB() { };
	inline ~StringB();
	inline StringB(const String&);
	inline StringB(const StringB& s) : StringB( (String&)s ) { }
	inline StringB(String&&);
	inline StringB(const wchar_t* data);
	inline StringB(const wchar_t* data, size_t len);
	StringB(const StringA&);

	inline String& operator=(const String& s);
	inline String& operator=(const StringB& s) { return (*this) = (String&) s; }
	inline String& operator=(StringB&& s);
	String& operator= (const StringA&);
	inline void operator+= (const String& s);
};

inline String::String() {
	tdata = nullptr;
	tlen  = 0;
}

inline String::~String() { }

inline StringB::~StringB() {
	if( tdata!=nullptr )
		heap_free(tdata);
	tdata = nullptr;
}

inline String::String(const String& s) {
	tdata = s.tdata;
	tlen  = s.tlen;
}

inline StringB::StringB(const String& s) : String() {
	if( s.tdata != nullptr ) {
		tdata = heap_alloc<wchar_t>(s.tlen+1);
		tlen = s.tlen;
		copy_array(s.tdata,s.tdata+(tlen+1),tdata);
	} else {
		tdata = nullptr;
		tlen  = 0;
	}
}

inline StringB::StringB(String&& s) {
	tdata   = s.tdata;
	tlen    = s.tlen;
	s.tdata = nullptr;
	s.tlen  = 0;
}

inline String::String(const wchar_t* data) {
	tdata = (wchar_t*) data;
	tlen  = wcslen(data);
}

inline StringB::StringB(const wchar_t* data) {
	tlen  = wcslen(data);
	tdata = heap_alloc<wchar_t>(tlen+1);
	copy_array(data,data+(tlen+1),tdata);
}

inline StringB::StringB(const wchar_t* data, size_t len) {
	tlen  = len;
	tdata = heap_alloc<wchar_t>(tlen+1);
	copy_array(data,data+tlen,tdata);
	tdata[tlen] = '\0';
}

inline String& String::operator=(const String& s) {
	tdata = s.tdata;
	tlen  = s.tlen;
	return *this;
}

inline String& StringB::operator=(const String& s) {
	if( tdata!=NULL )
		heap_free(tdata);
	if( s.tdata != NULL ) {
		tlen  = s.tlen;
		tdata = heap_alloc<wchar_t>(tlen+1);
		copy_array( s.tdata, s.tdata+(tlen+1), tdata );
	} else {
		tdata = NULL;
		tlen  = 0;
	}
	return *this;
}

inline String& StringB::operator=(StringB&& s) {
	if( tdata!=NULL )
		heap_free(tdata);
	tlen  = s.tlen;
	tdata = s.tdata;
	s.tlen  = 0;
	s.tdata = NULL;
	return *this;
}

inline StringB String::operator+ (const String& s) const {
	StringB res;
	res.tdata = heap_alloc<wchar_t>(tlen+s.tlen+1);
	res.tlen  = tlen+s.tlen;
	copy_array( tdata, tdata+(tlen+1), res.tdata );
	copy_array( s.tdata, s.tdata+(s.tlen+1), res.tdata+tlen );
	return res;
}

inline void StringB::operator+= (const String& s)  {
	*this = *this + s;
}

inline bool String::operator==(const String& s) const {
	if( s.tlen!=tlen )
		return false;
	for( size_t i=0; i<tlen; i++ )
		if( tdata[i] != s.tdata[i] )
			return false;
	return true;
}

inline bool String::operator==(const wchar_t* s) const {
	for( size_t i=0; ; i++ ) {
		if( tlen==i && s[i]=='\0' )
			return true;
		if( tdata[i] != s[i] )
			return false;
	}
}

inline bool String::operator!=(const String& s) const {
	return !( (*this) == s );
}

inline bool String::operator!=(const wchar_t* s) const {
	return !( (*this) == s );
}

inline bool String::operator< (const String& s) const {
	for( size_t i=0; ; i++ ) {
		if( tdata[i] < s.tdata[i] )
			return true;
		if( tdata[i] > s.tdata[i] )
			return false;
		if( tdata[i]=='\0' && s.tdata[i]=='\0' )
			return false;
	}
}

inline bool String::operator> (const String& s) const {
	for( size_t i=0; ; i++ ) {
		if( tdata[i] > s.tdata[i] )
			return true;
		if( tdata[i] < s.tdata[i] )
			return false;
		if( tdata[i]=='\0' && s.tdata[i]=='\0' )
			return false;
	}
}

inline bool String::operator<= (const String& s) const {
	return !( *this > s );
}

inline bool String::operator>= (const String& s) const {
	return !( *this < s );
}

inline void String::copy(wchar_t *buffer) const {
	copy_array(tdata,tdata+(tlen+1),buffer);
}


inline size_t String::strpos(wchar_t c) const {
	for( size_t i=0; i<tlen; i++ )
		if( tdata[i]==c )
			return i;
	return -1;
}

inline size_t String::strrpos(wchar_t c) const {
	for( size_t i=tlen-1; i>=0; i-- )
		if( tdata[i]==c )
			return i;
	return -1;
}

inline StringB String::substr(size_t begin) const {
	if( begin<0 || begin>tlen )
		throw 0;
	return StringB(tdata+begin,tlen-begin);
}

inline StringB String::substr(size_t begin, size_t length) const {
	if( begin<0 || begin+length>tlen )
		throw 0;
	return StringB(tdata+begin,length);
}


}