/******************************************
 **              CSR UTILS               **
 ** http://info.dcsr.ru/comp/csrutils/   **
 ******************************************
 **           Time-date utils            **
 ** Version: 0.1 (2015-05-24)            **
 ******************************************/

#pragma once

#include <csrutils.h>
#include <csrstr.h>

namespace csrutils {

class Time;
class TimeDiff;
class TimeZone;

struct TimeInfo {
	int year;
	int month;
	int day;
	int hour;
	int minute;
	int second;
	int millisecond;
};

class Time {
	uint64_t base;
public:
	inline Time() {
		base = 0;
	}
	inline Time(uint64_t _base) {
		base = _base;
	}

	Time(int year, int month, int day);
	Time(int year, int month, int day, int H, int M, int S);
	Time(TimeInfo);
	operator TimeInfo() const;
	operator String() const;

	inline TimeDiff operator-(const Time&) const;
	inline Time operator+(const TimeDiff& dif) const;
	inline void operator+=(const TimeDiff& dif);

	inline bool operator==(const Time& t) const { return base == t.base; }
	inline bool operator>(const Time& t) const { return base > t.base; }
	inline bool operator<(const Time& t) const { return base < t.base; }
	inline bool operator>=(const Time& t) const { return base >= t.base; }
	inline bool operator<=(const Time& t) const { return base <= t.base; }

	static Time fromUnix(uint32_t);
	static Time fromWindows(uint64_t);
#ifdef _WINDOWS_
	inline static Time fromWindows(FILETIME ft) {
		return fromWindows( *(uint64_t*)(&ft) );
	}
#endif
	uint32_t toUnix() const;

	TimeInfo toLocal() const;
	TimeInfo toLocal(const TimeZone&) const;
	static Time fromLocal(TimeInfo&);
	static Time fromLocal(TimeInfo&, const TimeZone&);

	String toLocalString() const;
	String toLocalString(const TimeZone&) const;

	inline static Time formBase(uint64_t b) { Time t; t.base = b; return t; }

	friend class TimeDiff;
};

class TimeDiff {
	bool use_info;
	int64_t base;
	TimeInfo info;

public:
	inline TimeDiff() {
		use_info = false;
		base = 0;
	}

	inline TimeDiff(int64_t _base) {
		use_info = false;
		base = _base;
	}

	inline TimeDiff(TimeInfo _info) {
		use_info = true;
		base = 0;
		info = _info;
	}

	inline static TimeDiff millisecond() {
		return TimeDiff(1L);
	}

	inline static TimeDiff second() {
		return TimeDiff(1000L);
	}

	inline static TimeDiff minute() {
		return TimeDiff(60000L);
	}

	inline static TimeDiff hour() {
		return TimeDiff(3600000L);
	}

	inline static TimeDiff day() {
		return TimeDiff(86400000L);
	}

	inline static TimeDiff week() {
		return TimeDiff(7*86400000L);
	}

	inline static TimeDiff month() {
		TimeInfo info = { 0, 1, 0, 0, 0, 0, 0 };
		return TimeDiff(info);
	}

	inline static TimeDiff year() {
		TimeInfo info = { 1, 0, 0, 0, 0, 0, 0 };
		return TimeDiff(info);
	}

	inline static TimeDiff century() {
		TimeInfo info = { 100, 0, 0, 0, 0, 0, 0 };
		return TimeDiff(info);
	}

	inline operator TimeInfo() const {
		if( use_info )
			return info;
		else
			return toComplex().info;
	}
	inline operator Time() const {
		return Time(base);
	}

	TimeDiff operator+(const TimeDiff&) const;
	inline TimeDiff operator-(const TimeDiff& d) const {
		return *this + (-d);
	}
	TimeDiff operator-() const;
	inline void operator+=(const TimeDiff& d) {
		*this = *this + d;
	}
	inline void operator-=(const TimeDiff& d) {
		*this = *this + d;
	}


	Time add(Time t) const;
	TimeDiff toComplex() const;
	TimeDiff toSimple() const;

};

inline Time Time::operator+(const TimeDiff& dif) const {
	return dif.add(*this);
}
inline void Time::operator+=(const TimeDiff& dif) {
	*this = dif.add(*this);
}

inline TimeDiff Time::operator-(const Time& t) const {
	return TimeDiff(base - t.base);
}

struct TimeZoneCtx;
class TimeZone {
	TimeZoneCtx* ctx;

public:
	TimeZone(int hours, int minutes);
	TimeZone(const TimeZone&);
	~TimeZone();

	String name() const;
	TimeDiff offsetUTC(Time) const;
	TimeDiff offsetLocal(Time) const;

	static TimeZone getZone(const String& name);
	static TimeZone getLocal();
};

};
