#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <ctype.h>

#include "csrutils.h"
#include "csrstream.h"
#include "csrxml.h"

#define BUF_SIZE		512

using namespace csrutils;

struct csrutils::XmlReaderCtx {
	char buf[BUF_SIZE];
	int buf_used;
	int buf_pos;
	int tok_state;
	XmlToken *tag;
	//int tag_npars;
	//char *tag_name;
	//char *tag_pars[16];
	//char *tag_vals[16];
};

#define TOK_TAGOPEN		1	// <
#define TOK_TAGCLOSE	2	// >
#define TOK_TAGNAME		3	// <key...	-> ket
#define TOK_PARAM		4	// par1="val1" -> par1
#define TOK_PARVAL		5	// par1="val1" -> val1
#define TOK_TAGIOPEN	6	// <?
#define TOK_TAGICLOSE	7	// ?>
#define TOK_TAGCLOSER	8	// /
#define TOK_TEXT		9	//

#define TOKST_BASIC		1	// *** <
#define TOKST_INTAG0	2	// tagname
#define TOKST_INTAG1	3	// par1
#define TOKST_INTAG2	4	// ="val1"

struct XMLTOK {
	int type;
	char *data;
};

int xml_nexttok(XmlReaderCtx *ctx, IOStream &input, XMLTOK* tok);
int xml_nextread(XmlReaderCtx *ctx, IOStream &input);
int xml_skipspace(XmlReaderCtx *ctx, IOStream &input);
void xml_trimbackspace(char *str);
bool xml_isspace(char c);
void xml_decodestr(char **str);

int xml_getstring(XmlReaderCtx *ctx, IOStream &input, char **string, int flags, char q=0);
#define GSTR_BR_SPACE	0x0001
#define GSTR_BR_EQUAL	0x0002
#define GSTR_BR_TAGCL	0x0004
#define GSTR_BR_TAGOP	0x0008
#define GSTR_BR_CUSTM	0x0010


int xml_readtag(XmlReaderCtx *ctx, IOStream &input, XMLTOK begin_tok);

XmlReader::XmlReader(IOStream &_input) : input(_input) {
	ctx = new XmlReaderCtx;
	ctx->buf_used  = 0;
	ctx->buf_pos   = 0;
	ctx->tok_state = TOKST_BASIC;
	XMLTOK tok;
	for(;;) {
		xml_nexttok(ctx,input,&tok);
		if( tok.type==TOK_TAGIOPEN )
			break;
		delete[] tok.data;
	}
	int ret = xml_readtag(ctx,input,tok);
	delete ctx->tag;	// TODO: store separatly xml-info, and delete in destructor
	ctx->tag = 0;
}

XmlReader::~XmlReader() {
	delete ctx;
}

XmlToken::~XmlToken() {
	if( text!=0 )
		delete[] text;
	for( int i=0; i<nparam; i++ ) {
		delete[] pars[i];
		delete[] vals[i];
	}
	if( pars!=0 )
		delete[] pars;
	if( vals!=0 )
		delete[] vals;
}

char *XmlToken::getValue(const char *key) {
	for(int i=0; i<nparam; i++) {
		if( !strcmp(pars[i],key) )
			return vals[i];
	}
	return 0;
}

XmlToken *XmlReader::getToken() {
	XMLTOK tok;
	int ret;
	ret = xml_nexttok(ctx,input,&tok);
	if( ret!=0 )
		return 0;
	if( tok.type == TOK_TAGOPEN ) {
		ret = xml_readtag(ctx,input,tok);
		if( ret!=0 )
			return 0;
		return ctx->tag;
	} if( tok.type == TOK_TEXT ) {
		XmlToken *tag = new XmlToken();
		tag->type   = XML_TEXT;
		tag->nparam = 0;
		tag->closed = false;
		tag->text   = tok.data;
		return tag;
	} //else {
		return 0;
	//}
}

int xml_readtag(XmlReaderCtx *ctx, IOStream &input, XMLTOK begin_tok) {
	XMLTOK tok;
	int tag_npars = 0;
	char *tag_pars[16];
	char *tag_vals[16];
	bool tag_closer = false;
	bool tag_closed = false;
	XmlToken *tag = new XmlToken();
	xml_nexttok(ctx,input,&tok);
	if( tok.type == TOK_TAGCLOSER ) {
		tag_closer = true;
		xml_nexttok(ctx,input,&tok);
	}
	if( tok.type != TOK_TAGNAME )
		return 2;
	tag->text = tok.data;
	for(;;) {
		xml_nexttok(ctx,input,&tok);
		if( tok.type == TOK_TAGCLOSE || tok.type == TOK_TAGICLOSE )
			break;
		else if( tok.type == TOK_PARAM ) {
			tag_pars[ tag_npars ] = tok.data;
			xml_nexttok(ctx,input,&tok);
			if( tok.type != TOK_PARVAL )
				return 2;
			tag_vals[ tag_npars ] = tok.data;
			tag_npars++;
		} else if( tok.type == TOK_TAGCLOSER )
			tag_closed = true;
		else
			return 2;
	}
	tag->type   = tag_closer ? XML_CLOSETAG : XML_OPENTAG;
	tag->nparam = tag_npars;
	tag->closed	= tag_closed;
	if( tag->nparam>0 ) {
		tag->pars = new char*[ tag->nparam ];
		tag->vals = new char*[ tag->nparam ];
	} else {
		tag->pars = 0;
		tag->vals = 0;
	}
	for( int i=0; i<tag->nparam; i++ ) {
		tag->pars[i] = tag_pars[i];
		tag->vals[i] = tag_vals[i];
	}
	ctx->tag = tag;
	return 0;
}

int xml_nexttok(XmlReaderCtx *ctx, IOStream &input, XMLTOK* tok) {
	if( ctx->buf_pos==ctx->buf_used )
		xml_nextread(ctx,input);
	if( ctx->buf_used == 0 )
		return 1;
	switch( ctx->tok_state ) {
	case TOKST_BASIC: {
		xml_skipspace(ctx,input);
		char c = ctx->buf[ ctx->buf_pos ];
		if( c=='<' ) {
			ctx->buf_pos++;
			if( ctx->buf_pos==ctx->buf_used )
				xml_nextread(ctx,input);
			char c2 = ctx->buf[ ctx->buf_pos ];
			if( c2=='?' ) {
				ctx->buf_pos ++;
				tok->type = TOK_TAGIOPEN;
				tok->data = NULL;
			} else {
				tok->type = TOK_TAGOPEN;
				tok->data = NULL;
			}
			ctx->tok_state = TOKST_INTAG0;
			return 0;
		} else {
			xml_getstring(ctx,input,&tok->data,GSTR_BR_TAGOP);
			xml_trimbackspace(tok->data);
			xml_decodestr(&tok->data);
			tok->type = TOK_TEXT;
			return 0;
		}
		break; }
	case TOKST_INTAG0: {
		xml_skipspace(ctx,input);
		if( ctx->buf_pos==ctx->buf_used )
			xml_nextread(ctx,input);
		char c = ctx->buf[ ctx->buf_pos ];
		if( c=='/' ) {
			ctx->buf_pos++;
			tok->type = TOK_TAGCLOSER;
			tok->data = NULL;
			return 0;
		}
		xml_getstring(ctx,input,&tok->data,GSTR_BR_SPACE|GSTR_BR_TAGCL);
		tok->type = TOK_TAGNAME;
		ctx->tok_state = TOKST_INTAG1;
		return 0; }
	case TOKST_INTAG1: {
		xml_skipspace(ctx,input);
		if( ctx->buf_pos==ctx->buf_used )
			xml_nextread(ctx,input);
		char c = ctx->buf[ ctx->buf_pos ];
		if( c=='?' ) {
			ctx->buf_pos++;
			if( ctx->buf_pos==ctx->buf_used )
				xml_nextread(ctx,input);
			ctx->buf_pos++;
			tok->type = TOK_TAGICLOSE;
			tok->data = NULL;
			ctx->tok_state = TOKST_BASIC;
			return 0;
		} else if( c=='>' ) {
			ctx->buf_pos++;
			tok->type = TOK_TAGCLOSE;
			tok->data = NULL;
			ctx->tok_state = TOKST_BASIC;
			return 0;
		} else if( c=='/' ) {
			ctx->buf_pos++;
			tok->type = TOK_TAGCLOSER;
			tok->data = NULL;
			return 0;
		}
		xml_getstring(ctx,input,&tok->data,GSTR_BR_EQUAL|GSTR_BR_TAGCL|GSTR_BR_SPACE);
		tok->type = TOK_PARAM;
		ctx->tok_state = TOKST_INTAG2;
		return 0; }
	case TOKST_INTAG2: {
		char c;
		xml_skipspace(ctx,input);
		if( ctx->buf_pos==ctx->buf_used )
			xml_nextread(ctx,input);
		c = ctx->buf[ ctx->buf_pos++ ];
		if( c!='=' )
			return 2;
		if( ctx->buf_pos==ctx->buf_used )
			xml_nextread(ctx,input);
		char q = ctx->buf[ ctx->buf_pos++ ];
		xml_getstring(ctx,input,&tok->data,GSTR_BR_CUSTM,q);
		xml_decodestr(&tok->data);
		if( ctx->buf_pos==ctx->buf_used )
			xml_nextread(ctx,input);
		c = ctx->buf[ ctx->buf_pos++ ];
		tok->type = TOK_PARVAL;
		ctx->tok_state = TOKST_INTAG1;
		return 0; }
	}
	return -1;
}

int xml_nextread(XmlReaderCtx *ctx, IOStream &input) {
	if( ctx->buf_pos==ctx->buf_used ) {
		size_t sz;
		int ret = input.read(ctx->buf,BUF_SIZE,sz);
		if( ret==IO_EOF && sz==0 ) {
			ctx->buf_used = 0;
			ctx->buf_pos  = 0;
			return 1;
		}
		ctx->buf_used = (int) sz;
		ctx->buf_pos  = 0;
		return 0;
	}
	return 0;
}

int xml_skipspace(XmlReaderCtx *ctx, IOStream &input) {
	for(;;) {
		if( ctx->buf_pos==ctx->buf_used )
			xml_nextread(ctx,input);
		if( ctx->buf_used==0 )
			return 0;
		char c = ctx->buf[ ctx->buf_pos ];
		if( !xml_isspace(c) )
			return 0;
		else
			ctx->buf_pos ++;
	}
}

int xml_getstring(XmlReaderCtx *ctx, IOStream &input, char **string, int flags, char q) {
	char *str = 0;
	int pos0 = ctx->buf_pos;
	for(;;) {
		if( ctx->buf_pos==ctx->buf_used ) {
			int len = 0;
			if( str!=0 )
				len = (int) strlen(str);
			char *str2 = new char[ len + ctx->buf_pos - pos0 + 1 ];
			if( str!=0 )
				strcpy(str2,str);
			memcpy(str2+len,ctx->buf+pos0,ctx->buf_pos-pos0);
			if( str!=0 )
				delete[] str;
			str2[ len+ctx->buf_pos-pos0 ] = '\0';
			str = str2;
			xml_nextread(ctx,input);
			if( ctx->buf_used==0 ) {
				*string = str;
				return 0;
			}
			pos0 = ctx->buf_pos;
		}
		char c = ctx->buf[ ctx->buf_pos++ ];
		if(
				( flags&GSTR_BR_SPACE ? xml_isspace(c) : false ) ||
				( flags&GSTR_BR_EQUAL ? c=='=' : false ) ||
				( flags&GSTR_BR_TAGCL ? c=='?' || c=='>' || c=='/' : false ) ||
				( flags&GSTR_BR_TAGOP ? c=='<' : false ) ||
				( flags&GSTR_BR_CUSTM ? c==q : false ) ) {
			ctx->buf_pos --;
			int len = 0;
			if( str!=0 )
				len = (int) strlen(str);
			char *str2 = new char[ len + ctx->buf_pos - pos0 + 1 ];
			if( str!=0 )
				strcpy(str2,str);
			memcpy(str2+len,ctx->buf+pos0,ctx->buf_pos-pos0);
			if( str!=0 )
				delete[] str;
			str2[ len+ctx->buf_pos-pos0 ] = '\0';
			str = str2;
			*string = str;
			return 0;
		}
	}
}

void xml_trimbackspace(char *str) {
	int i = (int) strlen(str) - 1;
	for( ; i>=0; i-- ) {
		if( !xml_isspace( str[i] ) ) {
			str[i+1] = '\0';
			return;
		}
	}
}

bool xml_isspace(char c) {
	return 0 != isspace( (unsigned char) c );
}

void xml_decodestr(char **str) {
	char *src = *str;
	char *dst = new char[ strlen(src) + 1 ];
	int i;
	int p=0;
	for( i=0; src[i]!='\0'; i++ ) {
		if( src[i]=='&' ) {
			i ++;
			char b[16];
			int j;
			for( j=0; j<15; j++ ) {
				if( src[i+j]==';' )
					break;
				b[j] = src[i+j];
			}
			b[j] = '\0';
			if( !strcmp(b,"lt") )
				dst[ p++ ] = '<';
			else if( !strcmp(b,"gt") )
				dst[ p++ ] = '>';
			else if( !strcmp(b,"amp") )
				dst[ p++ ] = '&';
			else if( !strcmp(b,"apos") )
				dst[ p++ ] = '\'';
			else if( !strcmp(b,"quot") )
				dst[ p++ ] = '\"';
			else
				dst[ p++ ] = '?';
			i += j;
		} else
			dst[ p++ ] = src[i];
	}
	dst[p] = '\0';
	*str = dst;
	delete[] src;
}
