/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **         IO stream utilities          **
 ** Version: 1.0 (2016 MARCH 27)         **
 ******************************************/

#include <list>
#include <csrutils.h>
#include <csrstream.h>
#include <csrmem.h>

using namespace csrutils;

struct ReaderStop {
	char *bytes;
	size_t size;

	bool operator== (ReaderStop& st) const {
		if( size != st.size )
			return false;
		return memcmp(bytes,st.bytes,size) == 0;
	}
};

struct csrutils::NestedStopReader_Ctx {
	std::list<ReaderStop,StdAllocator<ReaderStop>> stops;
	uint64_t start_position;
	uint64_t data_length;
	uint64_t position;
};

NestedStopReader::NestedStopReader(FileStream *_stream) : stream(_stream) {
	ctx = new(heap_alloc<NestedStopReader_Ctx>()) NestedStopReader_Ctx();
	ctx->start_position = stream->getposition();
	ctx->data_length = -1;
	ctx->position = 0;
}

NestedStopReader::~NestedStopReader() {
	for( auto iter=ctx->stops.begin(); iter!=ctx->stops.end(); iter++ )
		heap_free( (*iter).bytes );
	heap_free(ctx);
}

int NestedStopReader::write(const void *data, size_t size, size_t& written) { return 0; }
bool NestedStopReader::setend() { return false; }

// TODO: optimize implementation (this is dummy)
int NestedStopReader::read(void *data, size_t size, size_t& read) {
	size_t ureadsize = size;
	char *cdata = (char*) data;
	int ret = stream->read(data,ureadsize,read);
	for( auto iter=ctx->stops.begin(); iter!=ctx->stops.end(); iter++ ) {
		ReaderStop stop = *iter;
		size_t i, j;
		for( i=0; i<read; i++ ) {
			for( j=0; j<(read-i) && j<stop.size; j++ ) {
				if( cdata[i+j] != stop.bytes[j] )
					break;
			}
			if( j == stop.size ) {
				read = i;
				stream->setposition( ctx->start_position+ctx->position+read );
				ctx->position += read;
				return 0;
			}
			if( i+j==read ) {
				size_t sz = stop.size - j;
				char *buf = heap_alloc<char>(sz);
				size_t rsz;
				ret = stream->read(data,sz,rsz);
				size_t j0 = j;
				for(; j<rsz && j<stop.size && j<rsz; j++ ) {
					if( buf[j-j0] != stop.bytes[j] )
						break;
				}
				heap_free(buf);
				if( j == stop.size ) {
					read = i;
					stream->setposition( ctx->start_position+ctx->position+read );
					ctx->position += read;
					return 0;
				}
				stream->setposition( ctx->start_position+ctx->position+read );
			}
		}
	}
	ctx->position += read;
	return ret;
}

uint64_t NestedStopReader::getsize() {
	return ctx->data_length;
}

bool NestedStopReader::setsize(uint64_t size) {
	ctx->data_length = size;
	return true;
}

uint64_t NestedStopReader::getposition() {
	return ctx->position;
}

bool NestedStopReader::setposition(uint64_t pos) {
	ctx->position = pos;
	stream->setposition( ctx->start_position + pos );
	return true;
}

void NestedStopReader::setstop(const void *data, size_t size) {
	char* data2 = heap_alloc<char>(size);
	memcpy(data2,data,size);
	ReaderStop stop = { data2, size };
	ctx->stops.push_back(stop);
}

void NestedStopReader::unsetstop(const void *data, size_t size) {
	ReaderStop stop = { (char*)data, size };
	for( auto iter=ctx->stops.begin(); iter!=ctx->stops.end(); iter++ ) {
		if( (*iter) == stop ) {
			heap_free( (*iter).bytes );
			ctx->stops.erase(iter);
		}
	}
}

void NestedStopReader::setstop(const char *data) {
	setstop(data,strlen(data));
}

void NestedStopReader::unsetstop(const char *data) {
	unsetstop(data,strlen(data));
}