/******************************************
 **              CSR UTILS               **
 ** http://www.dcsr.ru/comp/csrutils/    **
 ******************************************
 **           Heap functions             **
 ** Version 1.1 (2016 APRIL 09)          **
 ******************************************/

#include <csrutils.h>
#include <csrmem.h>
#include <csrheap.h>
#include <csrosutils.h>
#include <csrmthread.h>

using namespace csrutils;

#define MBF_FREE	0x0001
#define MBF_LAST	0x0002

#define MB_SIGN		0x11112222	//'KCLB' = BLCK

#define ALIGN16(sz) (((((sz)-1)>>4)+1)<<4)

struct HEAP_BLOCK {
	uint32_t sign;
	uint32_t flags;
	//size_t used;
	size_t size;
	Heap_Ctx* heapctx;
};

struct HEAP_PAGE {
	HEAP_BLOCK *data;
	size_t offset;
	size_t size;
	HEAP_PAGE *next;
};

struct HEAP_ALLOCATOR {
	void* ctx;
	size_t pagesize;
	void* (*alloc)(void* ctx, size_t size);
	void (*release)(void* ctx, void* mem, size_t size);
};

struct csrutils::Heap_Ctx {
	uint64_t flags;
	HEAP_ALLOCATOR alloc;
	HEAP_PAGE *firtpage, *lastpage;
	SyncObject mutex;
};

HEAP_ALLOCATOR deafult_allocator();
HEAP_PAGE *alloc_page(Heap_Ctx *ctx, size_t size);
bool get_page_block_prev(Heap_Ctx *ctx, size_t offset, HEAP_PAGE*& page, HEAP_BLOCK*& block, HEAP_BLOCK*& prev);
bool join_with_prev(Heap_Ctx *ctx, HEAP_BLOCK*& block, HEAP_BLOCK *prev);
bool join_with_next(Heap_Ctx *ctx, HEAP_BLOCK *block);
bool block_split(Heap_Ctx *ctx, HEAP_BLOCK*& block, HEAP_BLOCK*& newblock, size_t size);

CSRLIB Heap_Ctx* heap_create(uint64_t flags) {
	Heap_Ctx* ctx = new Heap_Ctx;
	//Heap_Ctx* ctx = new( csrutils::heap_alloc<Heap_Ctx>() ) Heap_Ctx();
	ctx->flags    = flags;
	ctx->alloc    = deafult_allocator();
	ctx->firtpage = NULL;
	ctx->lastpage = NULL;
	return ctx;
}

CSRLIB void heap_destroy(Heap_Ctx* ctx) {
	HEAP_PAGE *page = ctx->firtpage;
	while(page) {
		ctx->alloc.release(ctx->alloc.ctx,page->data,page->size);
		HEAP_PAGE *next = page->next;
		delete page;
		page = next;
	}
	ctx->firtpage = NULL;
	ctx->lastpage = NULL;
	delete ctx;
	//csrutils::heap_free(ctx);
}

CSRLIB void* heap_alloc(Heap_Ctx* ctx, size_t size) {
	ctx->mutex.lock();
	HEAP_PAGE *page = ctx->firtpage;
	HEAP_BLOCK *block, *prevblock = NULL;
	for(;;) {
		if( page == NULL )
			page = alloc_page(ctx,size);
		block = page->data;
		prevblock = NULL;
		for(;;) {
			if( prevblock != NULL && (block->flags&MBF_FREE) && (prevblock->flags&MBF_FREE) ) {
				join_with_next(ctx,prevblock);
				block = prevblock;
			}
			if( (block->flags&MBF_FREE) && (block->size-sizeof(HEAP_BLOCK) >= size) )
				goto bfound;
			if( block->flags&MBF_LAST )
				break;
			prevblock = block;
			block = (HEAP_BLOCK*) ( (size_t)block + block->size );
		}
		page = page->next;
	}
bfound:
	HEAP_BLOCK* newblock;
	block_split(ctx,block,newblock,size);
	block->flags &= ~MBF_FREE;
	//block->used   = size;
	ctx->mutex.unlock();
	return (void*)( block + 1 );
}

CSRLIB void heap_free(Heap_Ctx* ctx, void* mem) {
	HEAP_BLOCK *block = (HEAP_BLOCK*) mem - 1;
	if( block->sign != MB_SIGN )
		throw 0;
	block->flags |= MBF_FREE;
}

CSRLIB void* heap_resize(Heap_Ctx* ctx, void* mem, size_t size) {
	// TODO:
	return mem;
}

HEAP_PAGE *alloc_page(Heap_Ctx *ctx, size_t size) {
	size_t pagesize = ( size > 0 ) ? ctx->alloc.pagesize*((size-1)/ctx->alloc.pagesize+1) : ctx->alloc.pagesize;
	HEAP_PAGE *page = (HEAP_PAGE*) ctx->alloc.alloc(ctx->alloc.ctx,pagesize);
	page->data = (HEAP_BLOCK*) ( page + 1 );
	page->data->sign    = MB_SIGN;
	page->data->flags   = MBF_FREE | MBF_LAST;
	//page->data->used    = 0;
	page->data->size    = pagesize - sizeof(HEAP_PAGE);
	page->data->heapctx = ctx;
	page->next   = NULL;
	page->size   = pagesize;
	//page->offset = (ctx->lastpage==NULL) ? 0 : (ctx->lastpage->offset+ctx->lastpage->size);
	if( ctx->firtpage == NULL ) {
		ctx->firtpage = page;
		ctx->lastpage = page;
	} else {
		ctx->lastpage->next = page;
		ctx->lastpage = page;
	}
	return page;
}

bool get_page_block_prev(Heap_Ctx *ctx, size_t offset, HEAP_PAGE*& _page, HEAP_BLOCK*& _block, HEAP_BLOCK*& _prev) {
	HEAP_PAGE *page = ctx->firtpage;
	while( page!=NULL ) {
		if( page->offset <= offset && offset < page->offset+page->size )
			break;
		page = page->next;
	}
	if( page==NULL )
		return false;
	HEAP_BLOCK *block = page->data;
	HEAP_BLOCK *prev = NULL;
	for(;;) {
		size_t off = (size_t)block - (size_t)page->data;
		if( off<=offset && offset<off+block->size )
			break;
		if( block->flags&MBF_LAST )
			return false;
		prev = block;
		block = (HEAP_BLOCK*) ( (size_t)block + block->size );
	}
	_page  = page;
	_block = block;
	_prev  = prev;
	return true;
}

bool join_with_prev(Heap_Ctx *ctx, HEAP_BLOCK*& block, HEAP_BLOCK *prev) {
	if( prev!=NULL && prev->flags&MBF_FREE ) {
		block->sign  = 0;
		prev->flags |= block->flags;
		prev->size   = prev->size + block->size;
		block = prev;
		return true;
	} else
		return false;
}

bool join_with_next(Heap_Ctx *ctx, HEAP_BLOCK *block) {
	if( !(block->flags&MBF_LAST) ) {
		HEAP_BLOCK *next = (HEAP_BLOCK*) ( (size_t)block + block->size );
		if( next->flags&MBF_FREE ) {
			next->sign = 0;
			block->flags |= next->flags;
			block->size   = block->size + next->size;
			return true;
		} else
			return false;
	} else
		return false;
}

bool block_split(Heap_Ctx *ctx, HEAP_BLOCK*& block, HEAP_BLOCK*& newblock, size_t size) {
	size_t newsize = ALIGN16( size + sizeof(HEAP_BLOCK) );
	if( block->size - newsize > sizeof(HEAP_BLOCK) ) {
		newblock = (HEAP_BLOCK*) ( (size_t)block + newsize );
		newblock->sign    = MB_SIGN;
		newblock->flags   = block->flags|MBF_FREE;
		newblock->size    = block->size - newsize;
		//newblock->used    = 0;
		newblock->heapctx = ctx;
		block->flags &= ~MBF_LAST;
		block->size   = newsize;
		join_with_next(ctx,newblock);
		return true;
	} else
		return false;
}

//#define DAFAULT_PAGESIZE	0x1000

void* default_alloc(void* ctx, size_t size) {
	return os_virtualalloc(size);
}

void default_release(void* ctx, void* mem, size_t size) {
	os_virtualfree(mem,size);
}

HEAP_ALLOCATOR deafult_allocator() {
	HEAP_ALLOCATOR alloc;
	alloc.ctx      = NULL;
	alloc.pagesize = MemorySection::getLargePageSize();
	alloc.alloc    = default_alloc;
	alloc.release  = default_release;
	return alloc;
}